""" MultiQC submodule to parse output from RSeQC junction_saturation.py
http://rseqc.sourceforge.net/#junction-saturation-py """

import logging
import re

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC junction_saturation frequency reports and parse their data"""

    # Set up vars
    self.junction_saturation_all = dict()
    self.junction_saturation_known = dict()
    self.junction_saturation_novel = dict()

    # Go through files and parse data
    for f in self.find_log_files("rseqc/junction_saturation"):
        parsed = dict()
        for line in f["f"].splitlines():
            r = re.search(r"^([xyzw])=c\(([\d,]+)\)$", line)
            if r:
                parsed[r.group(1)] = [float(i) for i in r.group(2).split(",")]
        if len(parsed) == 4:
            if parsed["z"][-1] == 0:
                log.warning(f"Junction saturation data all zeroes, skipping: '{f['s_name']}'")
            else:
                if f["s_name"] in self.junction_saturation_all:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="junction_saturation")
                self.junction_saturation_all[f["s_name"]] = dict()
                self.junction_saturation_known[f["s_name"]] = dict()
                self.junction_saturation_novel[f["s_name"]] = dict()
                for k, v in enumerate(parsed["x"]):
                    self.junction_saturation_all[f["s_name"]][v] = parsed["z"][k]
                    self.junction_saturation_known[f["s_name"]][v] = parsed["y"][k]
                    self.junction_saturation_novel[f["s_name"]][v] = parsed["w"][k]

    # Filter to strip out ignored sample names
    self.junction_saturation_all = self.ignore_samples(self.junction_saturation_all)
    self.junction_saturation_known = self.ignore_samples(self.junction_saturation_known)
    self.junction_saturation_novel = self.ignore_samples(self.junction_saturation_novel)

    if len(self.junction_saturation_all) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write data to file
    self.write_data_file(self.junction_saturation_all, "rseqc_junction_saturation_all")
    self.write_data_file(self.junction_saturation_known, "junction_saturation_known")
    self.write_data_file(self.junction_saturation_novel, "junction_saturation_novel")

    # Add line graph to section
    pconfig = {
        "id": "rseqc_junction_saturation_plot",
        "title": "RSeQC: Junction Saturation",
        "ylab": "Number of Junctions",
        "ymin": 0,
        "xlab": "Percent of reads",
        "xmin": 0,
        "xmax": 100,
        "tt_label": "<strong>{point.x}% of reads</strong>: {point.y:.2f}",
        "data_labels": [{"name": "All Junctions"}, {"name": "Known Junctions"}, {"name": "Novel Junctions"}],
        "cursor": "pointer",
        "click_func": "single_sample_plot",
    }
    self.add_section(
        name="Junction Saturation",
        anchor="rseqc-junction_saturation",
        description="""<a href="http://rseqc.sourceforge.net/#junction-saturation-py" target="_blank">Junction Saturation</a>
            counts the number of known splicing junctions that are observed
            in each dataset. If sequencing depth is sufficient, all (annotated) splice junctions should
            be rediscovered, resulting in a curve that reaches a plateau. Missing low abundance splice
            junctions can affect downstream analysis.</p>
            <div class="alert alert-info" id="rseqc-junction_sat_single_hint">
              <span class="glyphicon glyphicon-hand-up"></span>
              Click a line to see the data side by side (as in the original RSeQC plot).
            </div><p>""",
        plot=linegraph.plot(
            [self.junction_saturation_all, self.junction_saturation_known, self.junction_saturation_novel], pconfig
        ),
    )

    # Return number of samples found
    return len(self.junction_saturation_all)



""" MultiQC submodule to parse output from Rockhopper summary files
https://cs.wellesley.edu/~btjaden/Rockhopper/ """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialize the parent object
        super(MultiqcModule, self).__init__(
            name="Rockhopper",
            anchor="rockhopper",
            href="https://cs.wellesley.edu/~btjaden/Rockhopper/",
            info="""
            is a comprehensive and user-friendly system
            for computational analysis of bacterial RNA-seq data.
            It can align reads to genomes, assemble transcripts,
            identify transcript boundaries, and discover novel
            transcripts such as small RNAs.
            """,
            doi=["10.1016/j.ymeth.2019.03.026", "10.1186/s13059-014-0572-2", "10.1093/nar/gkt444"],
        )

        # Set up vars
        self.rh_data = dict()

        # Parse summary file
        for f in self.find_log_files("rockhopper"):
            self.parse_rockhopper_summary(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        # Filter to strip out ignored sample names
        self.rh_data = self.ignore_samples(self.rh_data)

        if len(self.rh_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.rh_data)} reports")

        # Write to file
        self.write_data_file(self.rh_data, "multiqc_rockhopper")

        # Basic stats Table
        self.rockhopper_general_stats_table()

        # Alignment bar plot
        self.rockhopper_count_bar_plot()

    def add_results_to_rhdata(self, f, s_name, results):
        """
        Helper function to add parsed results to rhdata
        """
        # Calculate unaligned read count
        total_mapped_reads = sum([v for k, v in results.items() if k != "total-reads"])
        results["unaligned"] = results["total-reads"] - total_mapped_reads

        if s_name in self.rh_data:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.add_data_source(f, s_name)
        self.rh_data[s_name] = results

    def parse_rockhopper_summary(self, f):
        s_name = None

        # Initialize stats fields
        stats_index = [
            "mRNA-sense",
            "mRNA-antisense",
            "rRNA-sense",
            "rRNA-antisense",
            "tRNA-sense",
            "tRNA-antisense",
            "ncRNA-sense",
            "ncRNA-antisense",
            "unannotated",
        ]

        results = {name: 0 for name in stats_index}
        # Files can have more than one sample in them. Store results for
        # each sample in a dictionary
        results_by_s_name = {}

        # Parse rockhopper output line-by-line since there may be many genomes
        lines = f["f"].split("\n")
        for i, line in enumerate(lines):
            # Get sample name
            if line.startswith("Aligning sequencing reads from file:"):
                # When reaching a new sample add the previous sample to
                # the dictionary of samples along with its results.
                # Then reset the results dictionary for the new sample
                if s_name and s_name not in results_by_s_name:
                    results_by_s_name[s_name] = results
                    results = {name: 0 for name in stats_index}

                s_name = line.split(":", 1)[1].strip()
                s_name = self.clean_s_name(s_name, f)

            elif line.startswith("Aligning sequencing reads from files:"):
                if s_name and s_name not in results_by_s_name:
                    results_by_s_name[s_name] = results
                    results = {name: 0 for name in stats_index}
                s_name = lines[i + 1].strip()
                s_name = self.clean_s_name(s_name, f)

            # Get total number of reads read by rockhopper
            if line.startswith("Total reads:"):
                results["total-reads"] = int(re.search(r"Total reads:\s*(\d*)", line).group(1))

            # Get number of reads aligned to each genome
            elif line.startswith("Successfully aligned reads"):
                # Get number of aligned reads
                genome_reads = int(re.search(r"Successfully aligned reads:\s*(\d*)", line).group(1))

                # Get percent of reads in each category
                stats = [int(re.search(r"(\d+)\%", subline).group(1)) for subline in lines[i + 1 : i + 10]]
                for name, val in zip(stats_index, stats):
                    # Convert percentages to true number of reads in each category
                    results[name] += int(round(val * genome_reads / 100))

        # Make sure the last sample name is added to the results dictionary
        if s_name and s_name not in results_by_s_name:
            results_by_s_name[s_name] = results

        # loop through all samples found and add them to rhdata
        for s_name in results_by_s_name:
            self.add_results_to_rhdata(f, s_name, results_by_s_name[s_name])

    def rockhopper_general_stats_table(self):
        """Take the parsed stats from the Rockhopper summary and add it to the
        basic stats table at the top of the report"""

        headers = {}
        headers["mRNA-sense"] = {
            "title": f"CDS Reads ({config.read_count_prefix})",
            "description": f"Reads aligned to coding regions ({config.read_count_desc})",
            "min": 0,
            "scale": "Blues",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
        }
        headers["mRNA-antisense"] = {
            "title": f"CDS Reads (a/s, {config.read_count_prefix})",
            "description": f"Antisense reads aligned to coding regions ({config.read_count_desc})",
            "min": 0,
            "scale": "Blues",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "hidden": True,
        }
        headers["rRNA-sense"] = {
            "title": f"rRNA Reads ({config.read_count_prefix})",
            "description": f"Reads aligned to rRNA ({config.read_count_desc})",
            "min": 0,
            "scale": "Blues",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
        }
        headers["rRNA-antisense"] = {
            "title": f"rRNA Reads (a/s, {config.read_count_prefix})",
            "description": f"Antisense reads aligned to rRNA ({config.read_count_desc})",
            "min": 0,
            "scale": "Blues",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "hidden": True,
        }
        self.general_stats_addcols(self.rh_data, headers)

    def rockhopper_count_bar_plot(self):
        """Stacked bar plot showing counts of reads"""

        pconfig = {
            "id": "rockhopper_reads_counts_plot",
            "title": "Rockhopper: Alignment types",
            "ylab": "Number of reads",
            "tt_percentage": False,
        }

        # Plot bar graph of groups
        keys = {
            "mRNA-sense": {"name": "mRNA (Sense)"},
            "mRNA-antisense": {"name": "mRNA (Antisense)"},
            "rRNA-sense": {"name": "rRNA (Sense)"},
            "rRNA-antisense": {"name": "rRNA (Antisense)"},
            "tRNA-sense": {"name": "tRNA (Sense)"},
            "tRNA-antisense": {"name": "tRNA (Antisense)"},
            "ncRNA-sense": {"name": "ncRNA (Sense)"},
            "ncRNA-antisense": {"name": "ncRNA (Antisense)"},
            "unannotated": {"name": "Unannotated"},
            "unaligned": {"name": "Unaligned"},
        }

        self.add_section(
            name="Rockhopper",
            anchor="rockhopper_reads_counts",
            description="""
            This plot shows the number of reads mapped to
            different regions of the genome, accounting for
            sense and antisense alignment, if relevant.
            """,
            plot=bargraph.plot(self.rh_data, keys, pconfig),
        )



""" MultiQC submodule to parse output from RSeQC read_duplication.py
http://rseqc.sourceforge.net/#read-duplication-py """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC read_duplication reports and parse their data"""

    # Set up vars
    self.read_dups = dict()

    # Go through files and parse data
    for f in self.find_log_files("rseqc/read_duplication_pos"):
        if f["f"].startswith("Occurrence	UniqReadNumber"):
            if f["s_name"] in self.read_dups:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="read_duplication")
            self.read_dups[f["s_name"]] = dict()
            for line in f["f"].splitlines():
                s = line.split()
                try:
                    if int(s[0]) <= 500:
                        self.read_dups[f["s_name"]][int(s[0])] = int(s[1])
                except Exception:
                    pass

    # Filter to strip out ignored sample names
    self.read_dups = self.ignore_samples(self.read_dups)

    if len(self.read_dups) == 0:
        return 0

    # Write data to file
    self.write_data_file(self.read_dups, "rseqc_read_dups")

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Add line graph to section
    pconfig = {
        "smooth_points": 200,
        "id": "rseqc_read_dups_plot",
        "title": "RSeQC: Read Duplication",
        "ylab": "Number of Reads (log10)",
        "xlab": "Occurrence of read",
        "yLog": True,
        "tt_label": "<strong>{point.x} occurrences</strong>: {point.y} reads",
    }

    self.add_section(
        name="Read Duplication",
        anchor="rseqc-read_dups",
        description='<a href="http://rseqc.sourceforge.net/#read-duplication-py" target="_blank">read_duplication.py</a>'
        " calculates how many alignment positions have a certain number of exact duplicates."
        " Note - plot truncated at 500 occurrences and binned.</p>",
        plot=linegraph.plot(self.read_dups, pconfig),
    )

    # Return number of samples found
    return len(self.read_dups)



from itertools import chain

from multiqc.plots import linegraph


def plot_qhist(samples, file_type, **plot_args):
    """Create line graph plot of histogram data for BBMap 'qhist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    sumy = sum([int(samples[sample]["data"][x][0]) for sample in samples for x in samples[sample]["data"]])

    cutoff = sumy * 0.999
    all_x = set()
    for item in sorted(chain(*[samples[sample]["data"].items() for sample in samples])):
        all_x.add(item[0])
        cutoff -= item[1][0]
        if cutoff < 0:
            xmax = item[0]
            break
    else:
        xmax = max(all_x)

    columns_to_plot = {
        "Linear": {0: "Read1", 3: "Read2"},
        "Logarithmic": {1: "Read1", 4: "Read2"},
        "Measured": {2: "Read1", 5: "Read2"},
    }

    plot_data = []
    for column_type in columns_to_plot:
        plot_data.append(
            {
                sample + "." + column_name: {
                    x: samples[sample]["data"][x][column] if x in samples[sample]["data"] else 0 for x in all_x
                }
                for sample in samples
                for column, column_name in columns_to_plot[column_type].items()
            }
        )

    plot_params = {
        "id": "bbmap-" + file_type + "_plot",
        "title": "BBTools: " + plot_args["plot_title"],
        "xmax": xmax,
        "xlab": "Position in read",
        "ylab": "Quality score",
        "data_labels": [
            {"name": "Linear", "ylab": "Quality score"},
            {"name": "Logarithmic", "ylab": "Log score"},
            {"name": "Measured", "ylab": "Measured quality value"},
        ],
    }

    plot_params.update(plot_args["plot_params"])
    plot = linegraph.plot(plot_data, plot_params)

    return plot



""" MultiQC functions to plot a bargraph """


import inspect
import logging
from collections import OrderedDict

import math
import re

from multiqc.utils import config, mqc_colour, report
from multiqc.plots.plotly import bar

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, cats=None, pconfig=None):
    """Plot a horizontal bar graph. Expects a 2D dict of sample
    data. Also, can take info about categories. There are quite a
    few variants of how to use this function, see the docs for details.
    :param data: 2D dict, first keys as sample names, then x:y data pairs
                 Can supply a list of dicts and will have buttons to switch
    :param cats: optional list or dict with plot categories
    :param pconfig: optional dict with config key:value pairs
    :return: HTML and JS, ready to be inserted into the page
    """

    if pconfig is None:
        pconfig = {}

    # Allow user to overwrite any given config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for k, v in config.custom_plot_config[pconfig["id"]].items():
            pconfig[k] = v

    # Validate config if linting
    if config.strict:
        # Get module name
        modname = ""
        callstack = inspect.stack()
        for n in callstack:
            if "multiqc/modules/" in n[1] and "base_module.py" not in n[1]:
                callpath = n[1].split("multiqc/modules/", 1)[-1]
                modname = f">{callpath}< "
                break
        # Look for essential missing pconfig keys
        for k in ["id", "title", "ylab"]:
            if k not in pconfig:
                errmsg = f"LINT: {modname}Bargraph pconfig was missing key '{k}'"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)
        # Check plot title format
        if not re.match(r"^[^:]*\S: \S[^:]*$", pconfig.get("title", "")):
            errmsg = "LINT: {} Bargraph title did not match format 'Module: Plot Name' (found '{}')".format(
                modname, pconfig.get("title", "")
            )
            logger.error(errmsg)
            report.lint_errors.append(errmsg)

    # Given one dataset - turn it into a list
    if not isinstance(data, list):
        data = [data]

    # Make list of cats from different inputs
    if cats is None:
        cats = list()
    elif not isinstance(cats, list):
        cats = [cats]
    elif cats and isinstance(cats[0], str):
        cats = [cats]
    # Generate default categories if not supplied
    for idx in range(len(data)):
        try:
            cats[idx]
        except IndexError:
            cats.append(list())
            for s in data[idx].keys():
                for k in data[idx][s].keys():
                    if k not in cats[idx]:
                        cats[idx].append(k)

    # If we have cats in lists, turn them into dicts
    for idx, cat in enumerate(cats):
        if isinstance(cat, list):
            newcats = dict()
            for c in cat:
                newcats[c] = {"name": c}
            cats[idx] = newcats
        else:
            for c in cat:
                if "name" not in cat[c]:
                    cats[idx][c]["name"] = c

    # Allow user to overwrite a given category config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for k, v in config.custom_plot_config[pconfig["id"]].items():
            for idx in range(len(cats)):
                if k in cats[idx].keys():
                    for kk, vv in v.items():
                        cats[idx][k][kk] = vv

    # Parse the data into a chart friendly format
    plotsamples = list()
    plotdata = list()
    for idx, d in enumerate(data):
        hc_samples = list(d.keys())
        if isinstance(d, OrderedDict):
            # Legacy: users assumed that passing an OrderedDict indicates that we
            # want to keep the sample order https://github.com/MultiQC/MultiQC/issues/2204
            pass
        elif pconfig.get("sort_samples", True):
            hc_samples = sorted(list(d.keys()))
        hc_data = list()
        sample_dcount = dict()
        for c in cats[idx].keys():
            thisdata = list()
            catcount = 0
            for s in hc_samples:
                if s not in sample_dcount:
                    sample_dcount[s] = 0

                if s not in d or c not in d[s]:
                    # Pad with NaNs when we have missing categories in a sample
                    thisdata.append(float("nan"))
                    continue
                val = d[s][c]
                if not isinstance(val, (float, int)):
                    try:
                        val = int(val)
                    except ValueError:
                        try:
                            val = float(val)
                        except ValueError:
                            val = None
                if val is None:
                    # Pad with NaNs when we have missing categories in a sample
                    thisdata.append(float("nan"))
                    continue
                if isinstance(val, float):
                    if math.floor(val) == val:
                        val = int(val)
                thisdata.append(val)
                catcount += 1
                sample_dcount[s] += 1
            if catcount > 0:
                if pconfig.get("hide_zero_cats", True) is False or max(x for x in thisdata if not math.isnan(x)) > 0:
                    thisdict = {"name": cats[idx][c]["name"], "data": thisdata}
                    if "color" in cats[idx][c]:
                        thisdict["color"] = cats[idx][c]["color"]
                    hc_data.append(thisdict)

        # Remove empty samples
        for s, c in sample_dcount.items():
            if c == 0:
                idx = hc_samples.index(s)
                del hc_samples[idx]
                for j, d in enumerate(hc_data):
                    del hc_data[j]["data"][idx]
        if len(hc_data) > 0:
            plotsamples.append(hc_samples)
            plotdata.append(hc_data)

    if len(plotdata) == 0:
        logger.warning(f"Tried to make bar plot, but had no data: {pconfig.get('id')}")
        return '<p class="text-danger">Error - was not able to plot data.</p>'

    # Add colors to the categories if not set. Since the "plot_defaults" scale is
    # identical to default scale of the Highcharts JS library, this is not strictly
    # needed. But it future proofs when we replace Highcharts with something else.
    scale = mqc_colour.mqc_colour_scale("plot_defaults")
    for si, sd in enumerate(plotdata):
        for di, d in enumerate(sd):
            d.setdefault("color", scale.get_colour(di, lighten=1))

    # Make a plot - custom, interactive or flat
    mod = get_template_mod()
    if "bargraph" in mod.__dict__ and callable(mod.bargraph):
        try:
            return mod.bargraph(plotdata, plotsamples, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return bar.plot(plotdata, plotsamples, pconfig)



""" MultiQC submodule to parse output from deepTools plotProfile """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class plotProfileMixin:
    def parse_plotProfile(self):
        """Find plotProfile output"""
        self.deeptools_plotProfile = dict()
        for f in self.find_log_files("deeptools/plotProfile", filehandles=False):
            parsed_data, bin_labels, converted_bin_labels = self.parsePlotProfileData(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotProfile:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotProfile[k] = v
            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotProfile")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.deeptools_plotProfile = self.ignore_samples(self.deeptools_plotProfile)

        if len(self.deeptools_plotProfile) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotProfile, "deeptools_plot_profile")

            # Try to do plot bands but don't crash if the labels aren't as we expect
            xPlotBands = []
            xPlotLines = []
            plotBandHelp = ""
            try:
                xPlotBands.append(
                    {
                        "from": converted_bin_labels[bin_labels.index("TES")],
                        "to": converted_bin_labels[-1],
                        "color": "#f7cfcf",
                    }
                )
                xPlotBands.append(
                    {
                        "from": converted_bin_labels[bin_labels.index("TSS")],
                        "to": converted_bin_labels[bin_labels.index("TES")],
                        "color": "#ffffe2",
                    }
                )
                xPlotBands.append(
                    {
                        "from": converted_bin_labels[0],
                        "to": converted_bin_labels[bin_labels.index("TSS")],
                        "color": "#e5fce0",
                    }
                )
                xPlotLines.append(
                    {
                        "width": 1,
                        "value": converted_bin_labels[bin_labels.index("TES")],
                        "dashStyle": "Dash",
                        "color": "#000000",
                    }
                )
                xPlotLines.append(
                    {
                        "width": 1,
                        "value": converted_bin_labels[bin_labels.index("TSS")],
                        "dashStyle": "Dash",
                        "color": "#000000",
                    }
                )
                plotBandHelp = """
                    * Green: {} upstream of gene to {}
                    * Yellow: {} to {}
                    * Pink: {} to {} downstream of gene
                    """.format(
                    list(filter(None, bin_labels))[0],
                    list(filter(None, bin_labels))[1],
                    list(filter(None, bin_labels))[1],
                    list(filter(None, bin_labels))[2],
                    list(filter(None, bin_labels))[2],
                    list(filter(None, bin_labels))[3],
                )
            except (ValueError, IndexError):
                pass

            config = {
                "id": "read_distribution_profile",
                "title": "deeptools: Read Distribution Profile after Annotation",
                "ylab": "Occurrence",
                "xlab": None,
                "smooth_points": 100,
                "xPlotBands": xPlotBands,
                "xPlotLines": xPlotLines,
            }

            self.add_section(
                name="Read Distribution Profile after Annotation",
                anchor="read_distribution_profile_plot",
                description="""
                    Accumulated view of the distribution of sequence reads related to the closest annotated gene.
                    All annotated genes have been normalized to the same size.

                    {}""".format(plotBandHelp),
                plot=linegraph.plot(self.deeptools_plotProfile, config),
            )

        return len(self.deeptools_plotProfile)

    def parsePlotProfileData(self, f):
        d = dict()
        bin_labels = []
        bins = []
        converted_bin_labels = []
        for line in f["f"].splitlines():
            cols = line.rstrip().split("\t")
            if cols[0] == "bin labels":
                for col in cols[2 : len(cols)]:
                    if col not in list(filter(None, bin_labels)):
                        bin_labels.append(col)
                    else:
                        break
            elif cols[0] == "bins":
                for col in cols[2 : len(cols)]:
                    if len(bins) != len(bin_labels):
                        bins.append(self._int(col))
                    else:
                        break
            else:
                s_name = self.clean_s_name(cols[0], f)
                d[s_name] = dict()

                # Convert the bins into genomic coordinates if we can
                try:
                    factors = {"Kb": 1e3, "Mb": 1e6, "Gb": 1e9}
                    convert_factor = 1
                    for k, v in factors.items():
                        if k in bin_labels[0]:
                            convert_factor *= v
                            start = float(bin_labels[0].strip(k)) * convert_factor
                    step = self._int(abs(start / bin_labels.index("TSS")))
                    end = step * (len(bin_labels) - bin_labels.index("TSS"))
                    converted_bin_labels = range((self._int(start) + step), (self._int(end) + step), step)
                except (UnboundLocalError, ValueError):
                    converted_bin_labels = bins

                for i in bins:
                    v = cols[i + 1]
                    try:
                        v = float(v)
                    except ValueError:
                        v = None
                    d[s_name].update({converted_bin_labels[i - 1]: v})

        return d, bin_labels, converted_bin_labels



""" MultiQC module to parse output from HISAT2 """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """HISAT2 module, parses stderr logs."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HISAT2",
            anchor="hisat2",
            href="https://ccb.jhu.edu/software/hisat2/",
            info="is a fast and sensitive alignment program for mapping "
            "NGS reads (both DNA and RNA) against a reference genome or "
            "population of reference genomes.",
            doi=["10.1038/nmeth.3317", "10.1038/s41587-019-0201-4"],
        )

        # Find and load any HISAT2 reports
        self.hisat2_data = dict()
        for f in self.find_log_files("hisat2", filehandles=True):
            self.parse_hisat2_logs(f)

        # Filter to strip out ignored sample names
        self.hisat2_data = self.ignore_samples(self.hisat2_data)

        if len(self.hisat2_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.hisat2_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None, f["s_name"])

        # Write parsed report data to a file
        self.write_data_file(self.hisat2_data, "multiqc_hisat2")

        # Basic Stats Table
        # Report table is immutable, so just updating it works
        self.hisat2_general_stats_table()

        # Alignment Rate Plot
        self.hisat2_alignment_plot()

    def parse_hisat2_logs(self, f):
        """
        Parse statistics generated by HISAT2 >= v2.1.0 that has been run
        with the --new-summary option. Older versions or logs from runs without
        that option are identical to that from bowtie2 and will be parsed
        by that module.
        """

        # Regexes
        regexes = {
            "unpaired_total": r"Total(?: unpaired)? reads: (\d+)",
            "unpaired_aligned_none": r"Aligned 0 times?: (\d+) \([\d\.]+%\)",
            "unpaired_aligned_one": r"Aligned 1 time: (\d+) \([\d\.]+%\)",
            "unpaired_aligned_multi": r"Aligned >1 times: (\d+) \([\d\.]+%\)",
            "paired_total": r"Total pairs: (\d+)",
            "paired_aligned_none": r"Aligned concordantly or discordantly 0 time: (\d+) \([\d\.]+%\)",
            "paired_aligned_one": r"Aligned concordantly 1 time: (\d+) \([\d\.]+%\)",
            "paired_aligned_multi": r"Aligned concordantly >1 times: (\d+) \([\d\.]+%\)",
            "paired_aligned_discord_one": r"Aligned discordantly 1 time: (\d+) \([\d\.]+%\)",
        }

        # Go through log file line by line
        s_name = f["s_name"]
        parsed_data = {}

        for line in f["f"]:
            # Attempt in vain to find original hisat2 command, logged by another program
            hscmd = re.search(r"hisat2 .+ -[1U] ([^\s,]+)", line)
            if hscmd:
                s_name = self.clean_s_name(hscmd.group(1), f)
                log.debug(f"Found a HISAT2 command, updating sample name to '{s_name}'")

            # Run through all regexes
            for k, r in regexes.items():
                match = re.search(r, line)
                if match:
                    parsed_data[k] = int(match.group(1))

            # Overall alignment rate
            overall = re.search(r"Overall alignment rate: ([\d\.]+)%", line)
            if overall:
                parsed_data["overall_alignment_rate"] = float(overall.group(1))

                # Save parsed data
                if s_name in self.hisat2_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.hisat2_data[s_name] = parsed_data

                # Reset in case we find more in this log file
                s_name = f["s_name"]
                parsed_data = {}

    def hisat2_general_stats_table(self):
        """Take the parsed stats from the HISAT2 report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "overall_alignment_rate": {
                "title": "% Aligned",
                "description": "overall alignment rate",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            }
        }
        self.general_stats_addcols(self.hisat2_data, headers)

    def hisat2_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        # Split the data into SE and PE
        sedata = {}
        pedata = {}
        for s_name, data in self.hisat2_data.items():
            if "paired_total" in data:
                # Save half 'pairs' of mate counts
                m_keys = ["unpaired_total", "unpaired_aligned_none", "unpaired_aligned_one", "unpaired_aligned_multi"]
                for k in m_keys:
                    if k in data:
                        data[k] = float(data[k]) / 2.0
                pedata[s_name] = data
            else:
                sedata[s_name] = data

        # Two plots, don't mix SE with PE
        if len(sedata) > 0:
            sekeys = {
                "unpaired_aligned_one": {"color": "#20568f", "name": "SE mapped uniquely"},
                "unpaired_aligned_multi": {"color": "#f7a35c", "name": "SE multimapped"},
                "unpaired_aligned_none": {"color": "#981919", "name": "SE not aligned"},
            }
            pconfig = {
                "id": "hisat2_se_plot",
                "title": "HISAT2: SE Alignment Scores",
                "ylab": "# Reads",
                "cpswitch_counts_label": "Number of Reads",
            }
            self.add_section(plot=bargraph.plot(sedata, sekeys, pconfig))

        if len(pedata) > 0:
            pekeys = {
                "paired_aligned_one": {"color": "#20568f", "name": "PE mapped uniquely"},
                "paired_aligned_discord_one": {"color": "#5c94ca", "name": "PE mapped discordantly uniquely"},
                "unpaired_aligned_one": {"color": "#95ceff", "name": "PE one mate mapped uniquely"},
                "paired_aligned_multi": {"color": "#f7a35c", "name": "PE multimapped"},
                "unpaired_aligned_multi": {"color": "#ffeb75", "name": "PE one mate multimapped"},
                "unpaired_aligned_none": {"color": "#981919", "name": "PE neither mate aligned"},
            }
            pconfig = {
                "id": "hisat2_pe_plot",
                "title": "HISAT2: PE Alignment Scores",
                "ylab": "# Reads",
                "cpswitch_counts_label": "Number of Reads",
            }
            self.add_section(
                description="<em>Please note that single mate alignment counts are halved to tally with pair counts properly.</em>",
                plot=bargraph.plot(pedata, pekeys, pconfig),
            )



""" MultiQC module to parse output from HiCExplorer """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HiCExplorer",
            anchor="hicexplorer",
            href="https://hicexplorer.readthedocs.io",
            info=" addresses the common tasks of Hi-C analysis from processing to visualization.",
            doi=["10.1093/nar/gky504", "10.1093/nar/gkaa220"],
        )

        self.hicexplorer_data = dict()
        for f in self.find_log_files("hicexplorer"):
            if f["fn"] != "QC_table.txt":
                # Parse the log file
                parsed_data = self.parse_logs(f["f"])
                # Build the sample ID
                s_name = f"{f['root']}_{f['s_name']}_{parsed_data['File']}"
                s_name = self.clean_s_name(s_name, f)
                # Save results
                if s_name in self.hicexplorer_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.hicexplorer_data[s_name] = parsed_data
                self.add_data_source(f, s_name=s_name)

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, s_name)

        self.hicexplorer_data = self.ignore_samples(self.hicexplorer_data)

        if len(self.hicexplorer_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.hicexplorer_data)} reports")

        self.write_data_file(self.hicexplorer_data, "multiqc_hicexplorer")

        self.colors = [
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#8c564b",
            "#7f7f7f",
            "#bcbd22",
            "#17becf",
            "#D2691E",
        ]

        # detect version of QC file
        # no 'Pairs mappable, unique and high quality' --> version 1.7
        # Contains 'pairs used' --> 1.8 - 3.1
        # Contains 'Hi-C contacts' --> since 3.2
        hicexplorer_versions = set()
        for s_name in self.hicexplorer_data:
            # compatibility to HiCExplorer <= 1.7 version QC files
            if (
                "Pairs mappable, unique and high quality" not in self.hicexplorer_data[s_name]
                and "Pairs considered" in self.hicexplorer_data[s_name]
            ):
                hicexplorer_versions.add("1.7")
                self.hicexplorer_data[s_name]["Pairs mappable, unique and high quality"] = self.hicexplorer_data[
                    s_name
                ]["Pairs considered"]
                for key in ["One mate unmapped", "One mate not unique", "One mate low quality"]:
                    self.hicexplorer_data[s_name]["Pairs mappable, unique and high quality"] -= self.hicexplorer_data[
                        s_name
                    ][key]

            # compatibility to HiCExplorer <= 3.1 version QC files
            if "Pairs considered" in self.hicexplorer_data[s_name]:
                self.hicexplorer_data[s_name]["Sequenced reads"] = self.hicexplorer_data[s_name]["Pairs considered"]
                self.hicexplorer_data[s_name]["Hi-c contacts"] = self.hicexplorer_data[s_name]["Pairs used"]
                self.hicexplorer_data[s_name]["Low mapping quality"] = self.hicexplorer_data[s_name][
                    "One mate low quality"
                ]
                self.hicexplorer_data[s_name]["Intra short range (< 20kb)"] = self.hicexplorer_data[s_name][
                    "Short range"
                ]
                self.hicexplorer_data[s_name]["Intra long range (>= 20kb)"] = self.hicexplorer_data[s_name][
                    "Long range"
                ]
                self.hicexplorer_data[s_name]["Read pair type: inward pairs"] = self.hicexplorer_data[s_name][
                    "Inward pairs"
                ]
                self.hicexplorer_data[s_name]["Read pair type: outward pairs"] = self.hicexplorer_data[s_name][
                    "Outward pairs"
                ]
                self.hicexplorer_data[s_name]["Read pair type: left pairs"] = self.hicexplorer_data[s_name][
                    "Left pairs"
                ]
                self.hicexplorer_data[s_name]["Read pair type: right pairs"] = self.hicexplorer_data[s_name][
                    "Right pairs"
                ]

            elif "Sequenced reads" in self.hicexplorer_data[s_name]:
                hicexplorer_versions.add("3.2")

        log.debug(f"HiCExplorer versions: {', '.join(hicexplorer_versions)}")

        # prepare the basic statistics for hicexplorer
        self.hicexplorer_basic_statistics()

        # key lists for plotting
        keys_categorization_of_reads_considered = [
            "Pairs mappable, unique and high quality",
            "One mate unmapped",
            "One mate not unique",
            "Low mapping quality",
        ]
        keys_mappable_unique_and_high_quality = [
            "Hi-c contacts",
            "Self ligation (removed)",
            "Same fragment",
            "Self circle",
            "Dangling end",
            "One mate not close to rest site",
            "Duplicated pairs",
        ]
        keys_list_contact_distance = ["Intra short range (< 20kb)", "Intra long range (>= 20kb)", "Inter chromosomal"]
        keys_list_read_orientation = [
            "Read pair type: inward pairs",
            "Read pair type: outward pairs",
            "Read pair type: left pairs",
            "Read pair type: right pairs",
            "Inter chromosomal",
        ]

        # prepare the detail report section
        self.add_section(
            name="Mapping statistics",
            anchor="hicexplorer_categorization_of_considered_reads",
            plot=self.hicexplorer_create_plot(
                keys_categorization_of_reads_considered,
                "HiCExplorer: Categorization of considered reads",
                "categorization",
            ),
            description="This shows how the sequenced read pairs were mapped and those filtered due to mapping problems.",
            helptext="""
                * **Pairs mappable, unique and high quality**
                    * The count of reads that were considered as valid reads and were not one of the following:
                * **One mate unmapped**
                    * Filtered out read because one mate was not mapped.
                * **One mate not unique**
                    * Filtered out read because one mate was not unique.
                * **Low mapping quality**
                    * Filtered out because one mate was having a low quality.
            """,
        )

        self.add_section(
            name="Read filtering",
            anchor="hicexplorer_pairs_categorized",
            plot=self.hicexplorer_create_plot(
                keys_mappable_unique_and_high_quality,
                "HiCExplorer: Categorization of reads - Pairs mappable, unique and high quality",
                "mapping",
            ),
            description="This figure contains the number of reads that were finally used to build the "
            "Hi-C matrix along with the reads that where filtered out.",
            helptext="""
            * **Dangling ends**
                * These are reads that start with the restriction site and constitute reads that were digested but no ligated.
            * **Same fragment**
                * These are read mates, facing inward, separated by up to 800 bp that do not have a restriction enzyme in between.
                * These read pairs are not valid Hi-C pairs.
            * **Self circle**
                * Self circles are defined as pairs within 25kb with 'outward' read orientation
            * **Self ligation**
                * These are read pairs with a restriction site in between that are within 800 bp.""",
        )

        self.add_section(
            name="Contact distance",
            anchor="hicexplorer_contact_distance",
            plot=self.hicexplorer_create_plot(
                keys_list_contact_distance, "HiCExplorer: Contact distance", "contact_distance"
            ),
            description="This figure contains information about the distance and location of the valid pairs used.",
            helptext="""
            * **Intra long range**
                * Pairs with a distance greater than 20 kilobases
            * **Intra short range**
                * Pairs with a distance less than 20 kilobases
            * **Inter chromosomal**
                * Interchromosomal pairs.
            """,
        )

        self.add_section(
            name="Read orientation",
            anchor="hicexplorer_read_orientation",
            plot=self.hicexplorer_create_plot(
                keys_list_read_orientation, "HiCExplorer: Read orientation", "orientation"
            ),
            description="This figure contains information about the orientation of the read pairs.",
            helptext="""
                * **Inward pairs**
                    * First mate is a forward read, second is reverse.
                    * `--------------->              <----------------`
                * **Outward pairs**
                    * First mate is a reverse read, second is forward.
                    * `--------------->              <----------------`
                * **Left pairs**
                    * Both are reverse reads.
                    * `<---------------              <----------------`
                * **Right pairs**
                    * Both are forward reads.
                    * `--------------->              ---------------->`
            """,
        )

    def parse_logs(self, f):
        """Parse a given HiCExplorer log file from hicBuildMatrix."""
        data = {}
        for line in f.splitlines():
            # catch empty lines
            if len(line) == 0:
                continue
            s = line.split("\t")
            # Skip header
            if s[0].startswith("#"):
                continue
            data_ = []
            # catch lines with descriptive content: "Of Hi-C contacts:"
            for i in s[1:]:
                if len(i) == 0:
                    continue
                try:
                    i = i.replace("(", "")
                    i = i.replace(")", "")
                    i = i.replace(",", "")
                    data_.append(float(i))
                except ValueError:
                    data_.append(i)
            if len(data_) == 0:
                continue
            if s[0].startswith("Intra short range (< 20kb)"):
                s[0] = "Intra short range (< 20kb)"
            elif s[0].startswith("same fragment"):
                s[0] = "same fragment"
            elif s[0].startswith("short range"):
                s[0] = "short range"
            s[0] = s[0].capitalize()
            data[s[0]] = data_[0]
            try:
                data[s[0] + "_pct"] = data_[1]
            except IndexError:
                pass
        return data

    def hicexplorer_basic_statistics(self):
        """Create the general statistics for HiCExplorer."""
        data = {}
        for s_name in self.hicexplorer_data:
            max_distance_key = "Max rest. site distance"
            total_pairs = self.hicexplorer_data[s_name]["Sequenced reads"]
            try:
                self.hicexplorer_data[s_name][max_distance_key]
            except KeyError:
                max_distance_key = "Max library insert size"
            data_ = {
                "Sequenced reads": self.hicexplorer_data[s_name]["Sequenced reads"],
                "Hi-c contacts": self.hicexplorer_data[s_name]["Hi-c contacts"] / total_pairs,
                "Mapped": self.hicexplorer_data[s_name]["One mate unmapped"] / total_pairs,
                "Min rest. site distance": self.hicexplorer_data[s_name]["Min rest. site distance"],
                max_distance_key: self.hicexplorer_data[s_name][max_distance_key],
            }
            data[s_name] = data_
        headers = {
            "Sequenced reads": {
                "title": f"{config.read_count_prefix} Pairs",
                "description": f"Total number of read pairs ({config.read_count_desc})",
                "shared_key": "read_count",
            },
            "Hi-c contacts": {
                "title": "% Used pairs",
                "max": 100,
                "min": 0,
                "modify": lambda x: x * 100,
                "suffix": "%",
            },
            "Mapped": {
                "title": "% Mapped",
                "max": 100,
                "min": 0,
                "modify": lambda x: (1 - x) * 100,
                "scale": "RdYlGn",
                "suffix": "%",
            },
            "Min rest. site distance": {
                "title": "Min RE dist",
                "description": "Minimum restriction site distance (bp)",
                "format": "{:.0f}",
                "suffix": " bp",
            },
            max_distance_key: {
                "title": "Max RE dist",
                "description": f"{max_distance_key} (bp)",
                "format": "{:.0f}",
                "suffix": " bp",
            },
        }

        self.general_stats_addcols(data, headers)

    def hicexplorer_create_plot(self, pKeyList, pTitle, pId):
        """Create the graphics containing information about the read quality."""

        keys = dict()
        for i, key_ in enumerate(pKeyList):
            keys[key_] = {"color": self.colors[i]}

        config = {
            "id": "hicexplorer_" + pId + "_plot",
            "title": pTitle,
            "ylab": "Number of Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.hicexplorer_data, keys, config)



""" MultiQC module to parse output from HOMER findpeaks """

import logging
import os

# Initialise the logger
log = logging.getLogger(__name__)


class FindPeaksReportMixin:
    def parse_homer_findpeaks(self):
        """Find HOMER findpeaks logs and parse their data"""

        self.homer_findpeaks = dict()
        for f in self.find_log_files("homer/findpeaks", filehandles=True):
            self.parse_find_peaks(f)

        # Filter to strip out ignored sample names
        self.homer_findpeaks = self.ignore_samples(self.homer_findpeaks)
        if not self.homer_findpeaks:
            return 0

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.homer_findpeaks, "multiqc_homer_findpeaks")

        # General Stats Table
        stats_headers = {
            "approximate_ip_efficiency": {
                "title": "% Efficiency",
                "description": "Approximate IP efficiency",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "total_peaks": {
                "title": "Total Peaks",
                "min": 0,
                "format": "{:,.0f}",
                "scale": "GnBu",
            },
            "expected_tags_per_peak": {
                "title": "Tags/Peak",
                "description": "Expected tags per peak",
                "min": 0,
                "format": "{:,.0f}",
                "scale": "PuRd",
            },
        }
        self.general_stats_addcols(self.homer_findpeaks, stats_headers, "findpeaks")

        return len(self.homer_findpeaks)

    def parse_find_peaks(self, f):
        """Parse HOMER findPeaks file headers."""
        parsed_data = dict()
        s_name = f["s_name"]
        for line in f["f"]:
            # Start of data
            if line.strip() and not line.strip().startswith("#"):
                break
            # Automatically parse header lines by = symbol
            s = line[2:].split("=")
            if len(s) > 1:
                k = s[0].strip().replace(" ", "_").lower()
                v = s[1].strip().replace("%", "")
                try:
                    parsed_data[k] = float(v)
                except ValueError:
                    parsed_data[k] = v
                if k == "tag_directory":
                    s_name = self.clean_s_name(os.path.basename(v), f, root=os.path.dirname(v))

        if len(parsed_data) > 0:
            if s_name in self.homer_findpeaks:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            self.add_data_source(f, s_name, section="findPeaks")
            self.homer_findpeaks[s_name] = parsed_data



#!/usr/bin/env python

""" MultiQC config module. Holds a single copy of
config variables to be used across all other modules """
from pathlib import Path
from typing import List, Dict, Optional, Union

import inspect

# Default logger will be replaced by caller
import logging
import os
import subprocess
import sys
from datetime import datetime

import importlib_metadata
import yaml
import pyaml_env

import multiqc
from multiqc.utils.util_functions import strtobool

logger = logging.getLogger("multiqc")

# Get the MultiQC version
version = importlib_metadata.version("multiqc_sgr")
short_version = version
git_hash = None
git_hash_short = None
script_path = str(Path(__file__).parent)  # dynamically used by util_functions.multiqc_dump_json()
git_root = None
try:
    git_root = subprocess.check_output(
        ["git", "rev-parse", "--show-toplevel"], cwd=script_path, stderr=subprocess.STDOUT, universal_newlines=True
    ).strip()
    git_root = Path(git_root)
    # .git
    # multiqc/
    #   utils/
    #       config.py  <- __file__
    expected_git_root = Path(script_path).parent.parent
    if git_root == expected_git_root:
        git_hash = subprocess.check_output(
            ["git", "rev-parse", "HEAD"], cwd=script_path, stderr=subprocess.STDOUT, universal_newlines=True
        ).strip()
        git_hash_short = git_hash[:7]
        version = f"{version} ({git_hash_short})"
except Exception:
    pass

# Constants
MULTIQC_DIR = os.path.dirname(os.path.realpath(inspect.getfile(multiqc)))

##### MultiQC Defaults
# Declaring variables for static typing
title: str
subtitle: str
intro_text: str
report_comment: str
report_header_info: List[Dict[str, str]]
show_analysis_paths: bool
show_analysis_time: bool
config_file: str
custom_logo: str
custom_logo_url: str
custom_logo_title: str
custom_css_files: List[str]
simple_output: bool
template: str
profile_runtime: bool
pandoc_template: str
read_count_multiplier: float
read_count_prefix: str
read_count_desc: str
long_read_count_multiplier: float
long_read_count_prefix: str
long_read_count_desc: str
base_count_multiplier: float
base_count_prefix: str
base_count_desc: str
output_fn_name: str
data_dir_name: str
plots_dir_name: str
data_format: str
force: bool
no_ansi: bool
quiet: bool
prepend_dirs: bool
prepend_dirs_depth: int
prepend_dirs_sep: str
file_list: bool
require_logs: bool
version_check_url: str

make_data_dir: bool
zip_data_dir: bool
data_dump_file: bool
megaqc_url: str
megaqc_access_token: str
megaqc_timeout: float
export_plots: bool
make_report: bool
plots_force_flat: bool
plots_force_interactive: bool
plots_flat_numseries: int
num_datasets_plot_limit: int
lineplot_style: str
lineplot_max_samples: int
barplot_legend_on_bottom: bool
violin_downsample_after: int
violin_min_threshold_outliers: int
violin_min_threshold_no_points: int
collapse_tables: bool
max_table_rows: int
table_columns_visible: Dict
table_columns_placement: Dict
table_columns_name: Dict
table_cond_formatting_colours: List[Dict[str, str]]
table_cond_formatting_rules: Dict[str, List[Dict[str, str]]]
decimalPoint_format: str
thousandsSep_format: str
remove_sections: List
section_comments: Dict
lint: bool  # Deprecated since v1.17
strict: bool
development: bool
custom_plot_config: Dict
custom_table_header_config: Dict
software_versions: Dict
ignore_symlinks: bool
ignore_images: bool
fn_ignore_dirs: List[str]
fn_ignore_paths: List[str]
sample_names_ignore: List[str]
sample_names_ignore_re: List[str]
sample_names_rename_buttons: List[str]
sample_names_replace: Dict
sample_names_replace_regex: bool
sample_names_replace_exact: bool
sample_names_replace_complete: bool
sample_names_rename: List
show_hide_buttons: List
show_hide_patterns: List
show_hide_regex: List
show_hide_mode: List
no_version_check: bool
log_filesize_limit: int
filesearch_lines_limit: int
report_readerrors: int
skip_generalstats: int
skip_versions_section: int
disable_version_detection: int
versions_table_group_header: str
data_format_extensions: Dict[str, str]
export_plot_formats: List[str]
filesearch_file_shared: List[str]
custom_content: Dict
fn_clean_sample_names: bool
use_filename_as_sample_name: bool
fn_clean_exts: List
fn_clean_trim: List
fn_ignore_files: List
top_modules: List
module_order: List[Union[str, Dict]]

# Populating the variables above from the default MultiQC config
config_defaults_path = os.path.join(MULTIQC_DIR, "utils", "config_defaults.yaml")
with open(config_defaults_path) as f:
    _default_config = yaml.safe_load(f)
for c, v in _default_config.items():
    globals()[c] = v

# Module filename search patterns
searchp_fn = os.path.join(MULTIQC_DIR, "utils", "search_patterns.yaml")
with open(searchp_fn) as f:
    sp = yaml.safe_load(f)

# Other defaults that can't be set in YAML
data_tmp_dir = "/tmp"  # will be overwritten by core script
modules_dir = os.path.join(MULTIQC_DIR, "modules")
creation_date = datetime.now().astimezone().strftime("%Y-%m-%d, %H:%M %Z")
working_dir = os.getcwd()
analysis_dir = [os.getcwd()]
output_dir = os.path.realpath(os.getcwd())
megaqc_access_token = os.environ.get("MEGAQC_ACCESS_TOKEN")

# Other variables that set only through the CLI
run_modules: List[str] = []
exclude_modules: List[str] = []
data_dir: Optional[str] = None
plots_tmp_dir: Optional[str] = None
plots_dir: Optional[str] = None
custom_data: Dict = {}
report_section_order: Dict = {}
output_fn: Optional[str] = None
megaqc_upload: bool = False

##### Available modules
# Modules must be listed in setup.py under entry_points['multiqc.modules.v1']
# Get all modules, including those from other extension packages
avail_modules = dict()
for entry_point in importlib_metadata.entry_points(group="multiqc.modules.v1"):
    nicename = entry_point.name
    avail_modules[nicename] = entry_point

##### Available templates
# Templates must be listed in setup.py under entry_points['multiqc.templates.v1']
# Get all templates, including those from other extension packages
avail_templates = {}
for entry_point in importlib_metadata.entry_points(group="multiqc.templates.v1"):
    nicename = entry_point.name
    avail_templates[nicename] = entry_point

##### Check we have modules & templates
# Check that we were able to find some modules and templates
# If not, package probably hasn't been installed properly.
# Need to do this before click, else will throw cryptic error
# Note: Can't use logger here, not yet initiated.
if len(avail_modules) == 0 or len(avail_templates) == 0:
    if len(avail_modules) == 0:
        print("Error - No MultiQC modules found.", file=sys.stderr)
    if len(avail_templates) == 0:
        print("Error - No MultiQC templates found.", file=sys.stderr)
    print(
        "Could not load MultiQC - has it been installed? \n\
        Please either install with pip (pip install multiqc) or by using \n\
        the local files (pip install .)",
        file=sys.stderr,
    )
    sys.exit(1)


##### Functions to load user config files. These are called by the main MultiQC script.
# Note that config files are loaded in a specific order and values can overwrite each other.
def mqc_load_userconfig(paths=()):
    """Overwrite config defaults with user config files"""

    # Load and parse installation config file if we find it
    mqc_load_config(os.path.join(os.path.dirname(MULTIQC_DIR), "multiqc_config.yaml"))

    # Load and parse a config file in $XDG_CONFIG_HOME
    # Ref: https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    mqc_load_config(
        os.path.join(os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~/.config")), "multiqc_config.yaml")
    )

    # Load and parse a user config file if we find it
    mqc_load_config(os.path.expanduser("~/.multiqc_config.yaml"))

    # Load and parse a config file path set in an ENV variable if we find it
    if os.environ.get("MULTIQC_CONFIG_PATH") is not None:
        mqc_load_config(os.environ.get("MULTIQC_CONFIG_PATH"))

    # Load separate config entries from MULTIQC_* environment variables
    mqc_add_config(mqc_env_vars_config())

    # Load and parse a config file in this working directory if we find it
    mqc_load_config("multiqc_config.yaml")

    # Custom command line config
    for p in paths:
        mqc_load_config(p)


def mqc_load_config(yaml_config_path: str):
    """Load and parse a config file if we find it"""
    if not os.path.isfile(yaml_config_path) and os.path.isfile(yaml_config_path.replace(".yaml", ".yml")):
        yaml_config_path = yaml_config_path.replace(".yaml", ".yml")

    if os.path.isfile(yaml_config_path):
        try:
            # pyaml_env allows referencing environment variables in YAML for default values
            new_config = pyaml_env.parse_config(yaml_config_path)
            logger.debug(f"Loading config settings from: {yaml_config_path}")
            mqc_add_config(new_config, yaml_config_path)
        except (IOError, AttributeError) as e:
            logger.debug(f"Config error: {e}")
        except yaml.scanner.ScannerError as e:
            logger.error(f"Error parsing config YAML: {e}")
            sys.exit(1)


def mqc_cl_config(cl_config):
    for clc_str in cl_config:
        try:
            parsed_clc = yaml.safe_load(clc_str)
            # something:var fails as it needs a space. Fix this (a common mistake)
            if isinstance(parsed_clc, str) and ":" in clc_str:
                clc_str = ": ".join(clc_str.split(":"))
                parsed_clc = yaml.safe_load(clc_str)
            assert isinstance(parsed_clc, dict)
        except yaml.scanner.ScannerError as e:
            logger.error(f"Could not parse command line config: {clc_str}\n{e}")
        except AssertionError:
            logger.error(f"Could not parse command line config: {clc_str}")
        else:
            logger.debug(f"Found command line config: {parsed_clc}")
            mqc_add_config(parsed_clc)


def mqc_env_vars_config() -> Dict:
    """
    Check MULTIQC_* environment variables and set to corresponding config values if they are of scalar types.
    """
    RESERVED_NAMES = {"MULTIQC_CONFIG_PATH"}
    PREFIX = "MULTIQC_"  # Prefix for environment variables
    env_config = {}
    for k, v in os.environ.items():
        if k.startswith(PREFIX) and k not in RESERVED_NAMES:
            conf_key = k[len(PREFIX) :].lower()
            if conf_key not in globals():
                continue
            if isinstance(globals()[conf_key], bool):
                try:
                    v = strtobool(v)
                except ValueError:
                    logger.warning(f"Could not parse a boolean value from the environment variable ${k}={v}")
                    continue
            elif isinstance(globals()[conf_key], int):
                try:
                    v = int(v)
                except ValueError:
                    logger.warning(f"Could not parse a int value from the environment variable ${k}={v}")
                    continue
            elif isinstance(globals()[conf_key], float):
                try:
                    v = float(v)
                except ValueError:
                    logger.warning(f"Could not parse a float value from the environment variable ${k}={v}")
                    continue
            elif not isinstance(globals()[conf_key], str) and globals()[conf_key] is not None:
                logger.warning(
                    f"Can only set scalar config entries (str, int, float, bool) with environment variable, "
                    f"but config.{conf_key} expects a type '{type(globals()[conf_key]).__name__}'. Ignoring ${k}"
                )
                continue
            env_config[conf_key] = v
            logger.debug(f"Setting config.{conf_key} from the environment variable ${k}")
    return env_config


def mqc_add_config(conf: Dict, conf_path=None):
    """Add to the global config with given MultiQC config dict"""
    global custom_css_files, fn_clean_exts, fn_clean_trim
    log_new_config = {}
    log_filename_patterns = []
    log_filename_clean_extensions = []
    log_filename_clean_trimmings = []
    for c, v in conf.items():
        if c == "sp":
            # Merge filename patterns instead of replacing
            sp.update(v)
            log_filename_patterns.append(v)
        elif c == "extra_fn_clean_exts":
            # Prepend to filename cleaning patterns instead of replacing
            fn_clean_exts[0:0] = v
            log_filename_clean_extensions.append(v)
        elif c == "extra_fn_clean_trim":
            # Prepend to filename cleaning patterns instead of replacing
            fn_clean_trim[0:0] = v
            log_filename_clean_trimmings.append(v)
        elif c in ["custom_logo"] and v:
            # Resolve file paths - absolute or cwd, or relative to config file
            fpath = v
            if os.path.exists(v):
                fpath = os.path.abspath(v)
            elif conf_path is not None and os.path.exists(os.path.join(os.path.dirname(conf_path), v)):
                fpath = os.path.abspath(os.path.join(os.path.dirname(conf_path), v))
            else:
                logger.error(f"Config '{c}' path not found, skipping ({fpath})")
                continue
            log_new_config[c] = fpath
            update({c: fpath})
        elif c == "custom_css_files":
            for fpath in v:
                if os.path.exists(fpath):
                    fpath = os.path.abspath(fpath)
                elif conf_path is not None and os.path.exists(os.path.join(os.path.dirname(conf_path), fpath)):
                    fpath = os.path.abspath(os.path.join(os.path.dirname(conf_path), fpath))
                else:
                    logger.error(f"CSS path '{c}' path not found, skipping ({fpath})")
                    continue
                logger.debug(f"Adding css file '{c}': {fpath}")
                if not custom_css_files:
                    custom_css_files = []
                custom_css_files.append(fpath)
        else:
            log_new_config[c] = v
            update({c: v})
    if len(log_new_config) > 0:
        logger.debug(f"New config: {log_new_config}")
    if len(log_filename_patterns) > 0:
        logger.debug(f"Added to filename patterns: {log_filename_patterns}")
    if len(log_filename_clean_extensions) > 0:
        logger.debug(f"Added to filename clean extensions: {log_filename_clean_extensions}")
    if len(log_filename_clean_trimmings) > 0:
        logger.debug(f"Added to filename clean trimmings: {log_filename_clean_trimmings}")


#### Function to load file containing a list of alternative sample-name swaps
# Essentially a fancy way of loading stuff into the sample_names_rename config var
# As such, can also be done directly using a config file
def load_sample_names(snames_file):
    global sample_names_rename_buttons, sample_names_rename
    num_cols = None
    try:
        with open(snames_file) as f:
            logger.debug(f"Loading sample renaming config settings from: {snames_file}")
            for line in f:
                s = line.strip().split("\t")
                if len(s) > 1:
                    # Check that we have consistent numbers of columns
                    if num_cols is None:
                        num_cols = len(s)
                    elif num_cols != len(s):
                        logger.warning(
                            "Inconsistent number of columns found in sample names file (skipping line): '{}'".format(
                                line.strip()
                            )
                        )
                    # Parse the line
                    if len(sample_names_rename_buttons) == 0:
                        sample_names_rename_buttons = s
                    else:
                        sample_names_rename.append(s)
                elif len(line.strip()) > 0:
                    logger.warning(f"Sample names file line did not have columns (must use tabs): {line.strip()}")
    except (IOError, AttributeError) as e:
        logger.error(f"Error loading sample names file: {e}")
    logger.debug(f"Found {len(sample_names_rename_buttons)} sample renaming patterns")


def load_replace_names(rnames_file):
    global sample_names_replace
    try:
        with open(rnames_file) as f:
            logger.debug(f"Loading sample replace config settings from: {rnames_file}")
            for line in f:
                s = line.strip().split("\t")
                if len(s) == 2:
                    sample_names_replace[s[0]] = s[1]
    except (IOError, AttributeError) as e:
        logger.error(f"Error loading sample names replacement file: {e}")
    logger.debug(f"Found {len(sample_names_replace)} sample replacing patterns")


def load_show_hide(sh_file):
    global show_hide_buttons, show_hide_patterns, show_hide_mode, show_hide_regex
    if sh_file:
        try:
            with open(sh_file, "r") as f:
                logger.debug(f"Loading sample renaming config settings from: {sh_file}")
                for line in f:
                    s = line.strip().split("\t")
                    if len(s) >= 3 and s[1] in ["show", "hide", "show_re", "hide_re"]:
                        show_hide_buttons.append(s[0])
                        show_hide_mode.append(s[1])
                        show_hide_patterns.append(s[2:])
                        show_hide_regex.append(s[1] not in ["show", "hide"])  # flag whether or not regex is turned on
        except AttributeError as e:
            logger.error(f"Error loading show patterns file: {e}")

    # Prepend a "Show all" button if we have anything
    # Do this outside of the file load block in case it was set in the config
    if len(show_hide_buttons) > 0:
        logger.debug(f"Found {len(show_hide_buttons)} show/hide patterns")
        show_hide_buttons.insert(0, "Show all")
        show_hide_mode.insert(0, "hide")
        show_hide_patterns.insert(0, [])
        show_hide_regex.insert(0, False)


# Keep track of all changes to the config
nondefault_config = dict()


def update(u):
    update_dict(nondefault_config, u)
    return update_dict(globals(), u)


def update_dict(target, source, none_only=False):
    """Recursively updates nested dict d from nested dict u"""
    for key, val in source.items():
        if isinstance(val, dict):
            target[key] = update_dict(target.get(key, {}), val)
        else:
            if not none_only or target.get(key) is None:
                if isinstance(val, list):
                    target[key] = val.copy()
                else:
                    target[key] = val
    return target



import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

from .util import average_pos_from_size

log = logging.getLogger(__name__)

N_QV = 2
ADAPTER_SEQS = ["AGATCGGAAGAG", "ATGGAATTCTCG", "CTGTCTCTTATA"]


class DragenReadMetrics(BaseMultiqcModule):
    """
    Rendering ALL THE THINGS!
    """

    def add_read_metrics(self):
        # Add each section in order
        self.per_seq_quality_plot()
        self.seq_length_dist_plot()

        return self.dragen_fastqc_data.keys()

    def per_seq_quality_plot(self):
        """Create the HTML for the per sequence quality score plot"""

        data = dict()
        GROUP = "READ MEAN QUALITY"
        MAX_QV = 64
        max_non_zero = 0
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()
                group_data = self.dragen_fastqc_data[s_name][mate][GROUP]
                for qv in range(MAX_QV):
                    metric = f"Q{qv} Reads"
                    count = group_data[metric]
                    if count > 0:
                        max_non_zero = max(qv, max_non_zero)
                    data[r_name][qv] = count

        for r_name in data:
            for qv in range(max_non_zero + 2, MAX_QV):
                del data[r_name][qv]

        if len(data) == 0:
            log.debug("per_seq_quality not found in DRAGEN FastQC reports")
            return None

        pconfig = {
            "id": "dragenqc_per_sequence_quality_scores_plot",
            "title": "DRAGEN-QC: Per-Sequence Quality Scores",
            "ylab": "Count",
            "xlab": "Mean Sequence Quality (Phred Quality Score)",
            "ymin": 0,
            "xmin": 0,
            "xDecimals": False,
            # 'colors': self.get_status_cols('per_sequence_quality_scores'),
            "tt_label": "<b>Phred {point.x}</b>: {point.y} reads",
            "xPlotBands": [
                {"from": 28, "to": 100, "color": "#c3e6c3"},
                {"from": 20, "to": 28, "color": "#e6dcc3"},
                {"from": 0, "to": 20, "color": "#e6c3c3"},
            ],
        }
        self.add_section(
            name="Per-Sequence Quality Scores",
            anchor="dragenqc_per_sequence_quality_scores",
            description="The number of reads with average quality scores. Shows if a subset of reads has poor quality.",
            helptext="""
            From the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/3%20Per%20Sequence%20Quality%20Scores.html):
            _The per sequence quality score report allows you to see if a subset of your
            sequences have universally low quality values. It is often the case that a
            subset of sequences will have universally poor quality, however these should
            represent only a small percentage of the total sequences._
            """,
            plot=linegraph.plot(data, pconfig),
        )

    def seq_length_dist_plot(self):
        """Create the HTML for the Sequence Length Distribution plot"""

        data = dict()
        seq_lengths = set()
        multiple_lenths = False
        avg_to_range = dict()
        GROUP = "READ LENGTHS"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()

                group_data = self.dragen_fastqc_data[s_name][mate][GROUP]
                for metric, value in group_data.items():
                    if value > 0:
                        avg_pos = average_pos_from_size(metric)
                        data[r_name][avg_pos] = value
                        avg_to_range[avg_pos] = metric.split("bp")[0]

                seq_lengths.update([avg_to_range[k] for k in data[r_name].keys()])

                if len(set(data[r_name].keys())) > 1:
                    multiple_lenths = True

        if len(data) == 0:
            log.debug("sequence_length_distribution not found in FastQC reports")
            return None

        if not multiple_lenths:
            lengths = "bp , ".join([str(line) for line in list(seq_lengths)])
            desc = f"All samples have sequences within a single length bin ({lengths}bp)."
            if len(seq_lengths) > 1:
                desc += ' See the <a href="#general_stats">General Statistics Table</a>.'
            self.add_section(
                name="Sequence Length Distribution",
                anchor="dragenqc_sequence_length_distribution",
                description=f'<div class="alert alert-info">{desc}</div>',
            )
        else:
            pconfig = {
                "id": "dragenqc_sequence_length_distribution_plot",
                "title": "DRAGEN-QC: Sequence Length Distribution",
                "ylab": "Read Count",
                "xlab": "Sequence Length (bp)",
                "ymin": 0,
                "yMinTickInterval": 0.1,
                "xDecimals": False,
                # 'colors': self.get_status_cols('sequence_length_distribution'),
                "tt_label": "<b>{point.x} bp</b>: {point.y}",
            }
            self.add_section(
                name="Sequence Length Distribution",
                anchor="dragenqc_sequence_length_distribution",
                description="""The distribution of fragment sizes (read lengths) found.
                    See the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/7%20Sequence%20Length%20Distribution.html)""",
                plot=linegraph.plot(data, pconfig),
            )



""" MultiQC module to parse output from THetA2 """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="THetA2",
            anchor="theta2",
            href="http://compbio.cs.brown.edu/projects/theta/",
            info="<em>(Tumor Heterogeneity Analysis)</em> estimates tumour purity "
            "and clonal / subclonal copy number.",
            doi=["10.1093/bioinformatics/btu651", "10.1186/gb-2013-14-7-r80"],
        )

        # Find and load any THetA2 reports
        self.theta2_data = dict()
        for f in self.find_log_files("theta2", filehandles=True):
            parsed_data = self.parse_theta2_report(f["f"])
            if len(parsed_data) > 0:
                if f["s_name"] in self.theta2_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f)
                self.theta2_data[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.theta2_data = self.ignore_samples(self.theta2_data)

        if len(self.theta2_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.theta2_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.theta2_data, "multiqc_theta2")

        # Alignment bar plot
        self.add_section(
            name="Tumour Subclone Purities",
            anchor="theta2-purities",
            description="Purities of tumour subclones. <em>NB:</em> Only first maximum likelihood solution for each sample shown.",
            plot=self.theta2_purities_chart(),
        )

    def parse_theta2_report(self, fh):
        """Parse the final THetA2 log file."""
        parsed_data = {}
        for line in fh:
            if line.startswith("#"):
                continue
            else:
                s = line.split("\t")
                purities = s[1].split(",")
                parsed_data["proportion_germline"] = float(purities[0]) * 100.0
                for i, v in enumerate(purities[1:]):
                    if i <= 5:
                        parsed_data[f"proportion_tumour_{i + 1}"] = float(v) * 100.0
                    else:
                        parsed_data["proportion_tumour_gt5"] = (float(v) * 100.0) + parsed_data.get(
                            "proportion_tumour_gt5", 0
                        )
                break
        return parsed_data

    def theta2_purities_chart(self):
        """Make the plot showing alignment rates"""

        # Specify the order of the different possible categories
        keys = {
            "proportion_germline": {"name": "Germline"},
            "proportion_tumour_1": {"name": "Tumour Subclone 1"},
            "proportion_tumour_2": {"name": "Tumour Subclone 2"},
            "proportion_tumour_3": {"name": "Tumour Subclone 3"},
            "proportion_tumour_4": {"name": "Tumour Subclone 4"},
            "proportion_tumour_5": {"name": "Tumour Subclone 5"},
            "proportion_tumour_gt5": {"name": "Tumour Subclones > 5"},
        }

        # Config for the plot
        pconfig = {
            "id": "theta2_purity_plot",
            "title": "THetA2: Tumour Subclone Purities",
            "cpswitch": False,
            "ymin": 0,
            "ymax": 100,
            "ylab": "% Purity",
            "tt_suffix": "%",
        }

        return bargraph.plot(self.theta2_data, keys, pconfig)



""" MultiQC module to parse output from Longranger """


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Longranger module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Long Ranger",
            anchor="longranger",
            href="https://support.10xgenomics.com/genome-exome/software/pipelines/latest/what-is-long-ranger",
            info="A set of analysis pipelines that perform sample demultiplexing, "
            "barcode processing, alignment, quality control, variant calling, phasing, "
            "and structural variant calling.",
            doi="10.1101/gr.234443.118",
        )

        def try_float_lambda(x, func, base):
            try:
                if func == "*":
                    return float(x) * base
                elif func == "/":
                    return float(x) / base
                else:
                    return x
            except Exception:
                return x

        self.headers = {
            "large_sv_calls": {
                "title": "Large SVs",
                "description": "Large structural variants called by Longranger. Not including blacklisted regions.",
                "format": "{:,.0f}",
                "scale": "PuRd",
            },
            "short_deletion_calls": {
                "title": "Short dels",
                "description": "Short deletions called by Longranger.",
                "format": "{:,.0f}",
                "scale": "PuRd",
                "hidden": True,
            },
            "genes_phased_lt_100kb": {
                "title": "genes phased < 100kb",
                "description": "Percentage of genes shorter than 100kb with >1 heterozygous SNP that are phased into a single phase block.",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "suffix": "%",
                "scale": "YlOrRd",
                "hidden": True,
            },
            "longest_phase_block": {
                "title": "Longest phased",
                "description": "Size of the longest phase block, in base pairs",
                "scale": "YlOrRd",
                "modify": lambda x: try_float_lambda(x, "/", 1000000.0),
                "suffix": "Mbp",
                "hidden": True,
            },
            "n50_phase_block": {
                "title": "N50 phased",
                "description": "N50 length of the called phase blocks, in base pairs.",
                "modify": lambda x: try_float_lambda(x, "/", 1000000.0),
                "suffix": "Mbp",
                "scale": "YlOrRd",
                "hidden": True,
            },
            "snps_phased": {
                "title": "SNPs phased",
                "description": "Percentage of called SNPs that were phased.",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "suffix": "%",
                "scale": "PuRd",
                "hidden": True,
            },
            "median_insert_size": {
                "title": "Insert size",
                "description": "Median insert size of aligned read pairs.",
                "format": "{:,.0f}",
                "suffix": "bp",
                "scale": "PuBu",
                "hidden": True,
            },
            "on_target_bases": {
                "title": "On target",
                "description": "Percentage of aligned bases mapped with the target regions in targeted mode. Only bases inside the intervals of target BED file are counted.",
                "suffix": "%",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "scale": "Greens",
            },
            "zero_coverage": {
                "title": "Zero cov",
                "description": "Percentage of non-N bases in the genome with zero coverage.",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "suffix": "%",
                "max": 100.0,
                "min": 0.0,
                "scale": "RdGy-rev",
            },
            "mean_depth": {
                "title": "Depth",
                "description": "Mean read depth, including PCR duplicate reads. In WGS mode, this is measured across the genome; in targeted mode, this is the measure inside targeted regions.",
                "suffix": "X",
                "scale": "PuBu",
            },
            "pcr_duplication": {
                "title": "PCR Dup",
                "description": "Percentage of reads marked as PCR duplicates. To be marked as PCR duplicates, reads must have the same mapping extents on the genome and the same 10x barcode.",
                "suffix": "%",
                "min": 15.0,
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "scale": "RdGy-rev",
                "hidden": True,
            },
            "mapped_reads": {
                "title": "Mapped",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "suffix": "%",
                "description": "Percentage of input reads that were mapped to the reference genome.",
                "scale": "PuBu",
                "hidden": True,
            },
            "number_reads": {
                "title": "M Reads",
                "modify": lambda x: try_float_lambda(x, "/", 1000000.0),
                "description": "Total number of reads supplied to Long Ranger. (millions)",
                "scale": "PuBu",
                "hidden": True,
            },
            "molecule_length_mean": {
                "title": "Mol size",
                "description": "The length-weighted mean input DNA length in base pairs.",
                "modify": lambda x: try_float_lambda(x, "/", 1000.0),
                "suffix": "Kbp",
                "scale": "YlGn",
            },
            "molecule_length_stddev": {
                "title": "Mol stddev",
                "description": "The length-weighted standard deviation of the input DNA length distribution in base pairs.",
                "modify": lambda x: try_float_lambda(x, "/", 1000.0),
                "suffix": "Kbp",
                "scale": "YlGn",
                "hidden": True,
            },
            "n50_linked_reads_per_molecule": {
                "title": "N50 read per mol.",
                "description": "The N50 number of read-pairs per input DNA molecule. Half of read-pairs came from molecules with this many or greater read-pairs.",
                "scale": "BuGn",
                "hidden": True,
            },
            "r1_q30_bases_fract": {
                "title": "% R1 >= Q30",
                "description": "Percentage of bases in R1 with base quality >= 30.",
                "hidden": True,
                "suffix": "%",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "scale": "Purples",
            },
            "r2_q30_bases_fract": {
                "title": "% R2 >= Q30",
                "description": "Percentage of bases in R2 with base quality >= 30.",
                "suffix": "%",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "scale": "Purples",
                "hidden": True,
            },
            "bc_on_whitelist": {
                "title": "Valid BCs",
                "description": "The Percentage of reads that carried a valid 10x barcode sequence.",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "suffix": "%",
                "scale": "BuPu",
                "hidden": True,
            },
            "bc_q30_bases_fract": {
                "title": "BC Q30",
                "description": "Percentage of bases in the barcode with base quality >= 30.",
                "suffix": "%",
                "modify": lambda x: try_float_lambda(x, "*", 100.0),
                "scale": "Purples",
                "hidden": True,
            },
            "bc_mean_qscore": {
                "title": "BC Qscore",
                "description": "The mean base quality value on the barcode bases.",
                "scale": "BuPu",
                "hidden": True,
            },
            "mean_dna_per_gem": {
                "title": "DNA per gem",
                "description": "The average number of base pairs of genomic DNA loaded into each GEM. This metric is based on the observed extents of read-pairs on each molecule.",
                "modify": lambda x: try_float_lambda(x, "/", 1000000.0),
                "suffix": "Mbp",
                "scale": "OrRd",
                "hidden": True,
            },
            "gems_detected": {
                "title": "M Gems",
                "description": "The number of Chromium GEMs that were collected and which generated a non-trivial number of read-pairs. (millions)",
                "modify": lambda x: try_float_lambda(x, "/", 1000000.0),
                "scale": "OrRd",
            },
            "corrected_loaded_mass_ng": {
                "title": "Loaded (corrected)",
                "description": "The estimated number of nanograms of DNA loaded into the input well of the Chromium chip. This metric is calculated by measuring the mean amount of DNA covered by input molecules in each GEM, then multiplying by the ratio of the chip input to the sample volume in each GEM.",
                "suffix": "ng",
                "scale": "RdYlGn",
            },
            "loaded_mass_ng": {
                "title": "Loaded",
                "description": "This metric was found to overestimate the true loading by a factor of 1.6, due primarily to denaturation of the input DNA.",
                "suffix": "ng",
                "scale": "RdYlGn",
            },
            "instrument_ids": {
                "title": "Instrument ID",
                "description": "The list of instrument IDs used to generate the input reads.",
                "scale": False,
                "hidden": True,
            },
            "longranger_version": {
                "title": "Long Ranger Version",
                "description": "The version of the Longranger software used to generate the results.",
                "scale": False,
            },
        }

        ### Parse the data
        self.longranger_data = dict()
        self.paths_dict = dict()
        for f in self.find_log_files("longranger/invocation"):
            sid = self.parse_invocation(f["f"])
            self.paths_dict[os.path.basename(f["root"])] = sid

        running_name = 1
        for f in self.find_log_files("longranger/summary"):
            data = self.parse_summary(f["f"])
            updir, _ = os.path.split(f["root"])
            base_updir = os.path.basename(updir)
            sid = f"longranger#{running_name}"
            if base_updir in self.paths_dict.keys():
                sid = self.paths_dict[base_updir]
            else:
                log.debug(f"Did not find _invocation file: {f['fn']}")
                running_name += 1

            self.longranger_data[sid] = data
            self.add_software_version(data["longranger_version"], sid)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.longranger_data = self.ignore_samples(self.longranger_data)

        if len(self.longranger_data) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.longranger_data.keys())} reports")

        # Write parsed report data to a file
        self.write_data_file(self.longranger_data, "multiqc_longranger")

        # Add a longranger versions column if not all the same
        longranger_versions = set([d["longranger_version"] for d in self.longranger_data.values()])
        version_str = ""
        if len(longranger_versions) == 1:
            version_str = f" All samples were processed using Longranger version {list(longranger_versions)[0]}"
            del self.headers["longranger_version"]

        ### Write the table
        config_table = {
            "id": "longranger_table",
            "namespace": "longranger",
            "title": "Long Ranger",
        }
        self.add_section(
            name="Run stats",
            anchor="longranger-run-stats",
            description="Statistics gathered from Longranger reports. "
            "There are more columns available but they are hidden by default." + version_str,
            helptext="""Parses the files `summary.csv` and `_invocation` found in the
                    output directory of Longranger. If `_invocation` is not found
                    the sample IDs will be missing and they will be given a running
                    number. E.g., `longranger#1` and `longranger#2`.""",
            plot=table.plot(self.longranger_data, self.headers, config_table),
        )

        ### Bar plot of phasing stats
        phase_pdata = {}
        snps_phased_pct = {}
        genes_phased_pct = {}
        for s_name in self.longranger_data:
            try:
                phase_pdata[s_name] = {
                    "longest_phase_block": float(self.longranger_data[s_name]["longest_phase_block"]),
                    "n50_phase_block": float(self.longranger_data[s_name]["n50_phase_block"]),
                }
            except Exception:
                pass
            try:
                snps_phased_pct[s_name] = {
                    "snps_phased_pct": float(self.longranger_data[s_name]["snps_phased"]) * 100.0
                }
            except Exception:
                pass
            try:
                genes_phased_pct[s_name] = {
                    "genes_phased_pct": float(self.longranger_data[s_name]["genes_phased_lt_100kb"]) * 100.0
                }
            except Exception:
                pass
        phase_plot_cats = [dict(), dict(), dict()]
        phase_plot_cats[0]["longest_phase_block"] = {"name": "Longest Phase Block"}
        phase_plot_cats[0]["n50_phase_block"] = {"name": "N50 of Phase Blocks"}
        phase_plot_cats[1]["snps_phased_pct"] = {"name": "% SNPs Phased"}
        phase_plot_cats[2]["genes_phased_pct"] = {"name": "% Genes < 100kbp in a single phase block"}
        if len(phase_pdata) > 0:
            self.add_section(
                name="Phasing",
                anchor="longranger-phasing",
                description="Phasing performance from Long Ranger. Genes are only considered if &le; 100kbp in length and with at least one heterozygous SNP.",
                helptext="""
                        * Longest phased
                            * Size of the longest phase block, in base pairs
                        * N50 phased
                            * N50 length of the called phase blocks, in base pairs.
                        * % SNPs phased
                            * Percentage of called SNPs that were phased.
                        * % Genes Phased
                            * Percentage of genes shorter than 100kb with >1 heterozygous SNP that are phased into a single phase block.
                        """,
                plot=bargraph.plot(
                    [phase_pdata, snps_phased_pct, genes_phased_pct],
                    phase_plot_cats,
                    {
                        "id": "longranger-phasing-plot",
                        "title": "Long Ranger: Phasing Statistics",
                        "data_labels": [
                            {"name": "N50 Phased", "ylab": "N50 of called phase blocks (bp)"},
                            {"name": "% SNPs Phased", "ylab": "% SNPs Phased", "ymax": 100},
                            {"name": "% Genes Phased", "ylab": "% Genes Phased", "ymax": 100},
                        ],
                        "cpswitch": False,
                        "stacking": None,
                        "ylab": "N50 of called phase blocks (bp)",
                    },
                ),
            )

        ### Bar plot of mapping statistics
        mapping_counts_data = {}
        for s_name in self.longranger_data:
            mapped_reads = float(self.longranger_data[s_name]["number_reads"]) * float(
                self.longranger_data[s_name]["mapped_reads"]
            )
            unmapped_reads = float(self.longranger_data[s_name]["number_reads"]) - mapped_reads
            dup_reads = mapped_reads * float(self.longranger_data[s_name]["pcr_duplication"])
            unique_reads = mapped_reads - dup_reads
            mapping_counts_data[s_name] = {
                "unique_reads": unique_reads,
                "dup_reads": dup_reads,
                "unmapped_reads": unmapped_reads,
            }
        mapping_counts_cats = dict()
        mapping_counts_cats["unique_reads"] = {"name": "Uniquely Aligned Reads", "color": "#437bb1"}
        mapping_counts_cats["dup_reads"] = {"name": "PCR Duplicate Aligned Reads", "color": "#7cb5ec"}
        mapping_counts_cats["unmapped_reads"] = {"name": "Unaligned Reads", "color": "#7f0000"}
        self.add_section(
            name="Alignment",
            anchor="longranger-alignment",
            description="Long Ranger alignment against the reference genome. To be marked as PCR duplicates, reads must have the same mapping extents on the genome and the same 10x barcode.",
            plot=bargraph.plot(
                mapping_counts_data,
                mapping_counts_cats,
                {
                    "id": "longranger-alignment-plot",
                    "title": "Long Ranger: Alignment Statistics",
                    "ylab": "Reads Counts",
                    "cpswitch_counts_label": "Read Counts",
                },
            ),
        )

    @staticmethod
    def parse_invocation(content):
        sid_pat = re.compile('    sample_id = "(.*)",')

        sid = None
        for line in content.splitlines():
            sid_m = re.match(sid_pat, line)
            if sid_m is not None:
                sid = sid_m.groups()[0]
        return sid

    @staticmethod
    def parse_summary(content):
        out_dict = dict()
        lines = content.splitlines()
        data = list(zip(lines[0].strip().split(","), lines[1].strip().split(",")))
        for i, j in data:
            if j != "":
                try:
                    out_dict[i] = float(j)
                except ValueError:
                    out_dict[i] = j

        return out_dict



import json

# Initialise the logger
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph
from multiqc.utils import report

from .util import average_from_range, average_pos_from_metric

log = logging.getLogger(__name__)

N_QV = 2
ADAPTER_SEQS = ["AGATCGGAAGAG", "ATGGAATTCTCG", "CTGTCTCTTATA"]


class DragenContentMetrics(BaseMultiqcModule):
    """
    Rendering ALL THE THINGS!
    """

    def add_content_metrics(self):
        # Now add each section in order
        self.n_content_plot()
        self.sequence_content_plot()
        self.adapter_content_plot()

        return self.dragen_fastqc_data.keys()

    def n_content_plot(self):
        """Create the HTML for the per base N content plot"""
        data = dict()
        GROUP = "POSITIONAL BASE CONTENT"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                totals = defaultdict(int)
                non_n = defaultdict(int)

                # Count total bases
                total_group_data = self.dragen_fastqc_data[s_name][mate][GROUP]
                for metric, value in total_group_data.items():
                    avg_pos = average_pos_from_metric(metric)
                    totals[avg_pos] += value
                    base = metric.split()[2]
                    if base != "N":
                        non_n[avg_pos] += value

                # Convert Total and Non-N counts into N%
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()
                for pos, count in totals.items():
                    if count == 0:
                        continue
                    non_n_count = non_n[pos]
                    n_count = count - non_n_count
                    n_frac = 100.0 * n_count / float(count)
                    data[r_name][pos] = n_frac

        if len(data) == 0:
            log.debug("per_base_n_content not found in DRAGEN FastQC reports")
            return None

        pconfig = {
            "id": "dragenqc_per_base_n_content_plot",
            "title": "DRAGEN-QC: Per-Position N Content",
            "ylab": "Percentage N-Count",
            "xlab": "Position in Read (bp)",
            "yCeiling": 100,
            "yMinRange": 5,
            "ymin": 0,
            "xmin": 0,
            "xDecimals": False,
            # 'colors': self.get_status_cols('per_base_n_content'),
            "tt_label": "<b>Base {point.x}</b>: {point.y:.2f}%",
            "yPlotBands": [
                {"from": 20, "to": 100, "color": "#e6c3c3"},
                {"from": 5, "to": 20, "color": "#e6dcc3"},
                {"from": 0, "to": 5, "color": "#c3e6c3"},
            ],
        }

        self.add_section(
            name="Per-Position N Content",
            anchor="dragenqc_per_base_n_content",
            description="The percentage of base calls at each position for which an `N` was called.",
            helptext="""
            From the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/6%20Per%20Base%20N%20Content.html):
            _If a sequencer is unable to make a base call with sufficient confidence then it will
            normally substitute an `N` rather than a conventional base call. This graph shows the
            percentage of base calls at each position for which an `N` was called._
            _It's not unusual to see a very low proportion of Ns appearing in a sequence, especially
            nearer the end of a sequence. However, if this proportion rises above a few percent
            it suggests that the analysis pipeline was unable to interpret the data well enough to
            make valid base calls._
            """,
            plot=linegraph.plot(data, pconfig),
        )

    def sequence_content_plot(self):
        """Create the epic HTML for the FastQC sequence content heatmap"""

        # Prep the data
        data = dict()
        GROUP = "POSITIONAL BASE CONTENT"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()
                group_data = self.dragen_fastqc_data[s_name][mate][GROUP]

                totals = defaultdict(int)
                for metric, value in group_data.items():
                    parts = metric.split()
                    # avg_pos = average_from_range(parts[1])
                    avg_pos = parts[1]
                    base = parts[-2].lower()

                    if avg_pos not in data[r_name]:
                        data[r_name][avg_pos] = dict()

                    # Store the current count and add it to the total
                    data[r_name][avg_pos][base] = value
                    totals[avg_pos] += value

                # Use the accumulated totals to normalize each bin to a percentage
                for pos, total in totals.items():
                    if total == 0:
                        del data[r_name][pos]
                        continue
                    for base in "acgt":
                        try:
                            data[r_name][pos][base] = (float(data[r_name][pos][base]) / float(total)) * 100.0
                        except Exception:
                            pass
                    data[r_name][pos]["base"] = pos

        if len(data) == 0:
            log.debug("sequence_content not found in FastQC reports")
            return None

        html = """<div id="dragen_fastqc_per_base_sequence_content_plot_div">
            <div class="alert alert-info">
               <span class="glyphicon glyphicon-hand-up"></span>
               Click a sample row to see a line plot for that dataset.
            </div>
            <h5><span class="s_name text-primary"><span class="glyphicon glyphicon-info-sign"></span> Rollover for sample name</span></h5>
            <div class="fastqc_seq_heatmap_key">
                Position: <span id="fastqc_seq_heatmap_key_pos">-</span>
                <div><span id="fastqc_seq_heatmap_key_t"> %T: <span>-</span></span></div>
                <div><span id="fastqc_seq_heatmap_key_c"> %C: <span>-</span></span></div>
                <div><span id="fastqc_seq_heatmap_key_a"> %A: <span>-</span></span></div>
                <div><span id="fastqc_seq_heatmap_key_g"> %G: <span>-</span></span></div>
            </div>
            <div id="fastqc_seq_heatmap_div" class="fastqc-overlay-plot">
                <div id="{id}" class="dragen_fastqc_per_base_sequence_content_plot hc-plot has-custom-export">
                    <canvas id="fastqc_seq_heatmap" height="100%" width="800px" style="width:100%;"></canvas>
                </div>
            </div>
            <div class="clearfix"></div>
        </div>
        <script type="application/json" class="fastqc_seq_content">{d}</script>
        """.format(
            # Generate unique plot ID, needed in mqc_export_selectplots
            id=report.save_htmlid("dragen_fastqc_per_base_sequence_content_plot"),
            d=json.dumps([self.anchor.replace("-", "_"), data]),
        )

        self.add_section(
            name="Per-Position Sequence Content",
            anchor="dragen_fastqc_per_base_sequence_content",
            description="The proportion of each base position for which each of the four normal DNA bases has been called.",
            helptext="""
            To enable multiple samples to be shown in a single plot, the base composition data
            is shown as a heatmap. The colours represent the balance between the four bases:
            an even distribution should give an even muddy brown colour. Hover over the plot
            to see the percentage of the four bases under the cursor.
            **To see the data as a line plot, as in the original FastQC graph, click on a sample track.**
            From the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/4%20Per%20Base%20Sequence%20Content.html):
            _Per Base Sequence Content plots out the proportion of each base position in a
            file for which each of the four normal DNA bases has been called._
            _In a random library you would expect that there would be little to no difference
            between the different bases of a sequence run, so the lines in this plot should
            run parallel with each other. The relative amount of each base should reflect
            the overall amount of these bases in your genome, but in any case they should
            not be hugely imbalanced from each other._
            _It's worth noting that some types of library will always produce biased sequence
            composition, normally at the start of the read. Libraries produced by priming
            using random hexamers (including nearly all RNA-Seq libraries) and those which
            were fragmented using transposases inherit an intrinsic bias in the positions
            at which reads start. This bias does not concern an absolute sequence, but instead
            provides enrichement of a number of different K-mers at the 5' end of the reads.
            Whilst this is a true technical bias, it isn't something which can be corrected
            by trimming and in most cases doesn't seem to adversely affect the downstream
            analysis._
            """,
            content=html,
        )

    def adapter_content_plot(self):
        """Create the epic HTML for the FastQC adapter content plot"""

        # Prep the data
        data = dict()
        COUNT_GROUP = "POSITIONAL BASE CONTENT"
        ADP_GROUP = "SEQUENCE POSITIONS"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                totals = defaultdict(int)
                for key, value in self.dragen_fastqc_data[s_name][mate][COUNT_GROUP].items():
                    parts = key.split()
                    pos = average_from_range(parts[1])
                    totals[pos] += int(value)

                adps = defaultdict(int)
                for key, value in self.dragen_fastqc_data[s_name][mate][ADP_GROUP].items():
                    parts = key.split()
                    seq = parts[0].split("'")[1]
                    if seq not in ADAPTER_SEQS:
                        continue
                    # avoid issues with metrics like "'AGATCGGAAGAG' Total Sequence Starts" where
                    # code attempts to parse 'Tot' as an integer
                    if not parts[1][:-2].isnumeric():
                        continue
                    pos = average_from_range(parts[1][:-2])
                    adps[pos] += int(value)

                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()
                cumsum = 0
                for pos, adp_count in sorted(adps.items()):
                    total = totals[pos]
                    cumsum += adp_count
                    if total > 0 and cumsum > 0:
                        data[r_name][pos] = 100.0 * cumsum / total

        pconfig = {
            "id": "dragen_fastqc_adapter_content_plot",
            "title": "FastQC: Adapter Content",
            "ylab": "% of Sequences",
            "xlab": "Position (bp)",
            "yCeiling": 100,
            "yMinRange": 5,
            "ymin": 0,
            "xDecimals": False,
            "tt_label": "<b>Base {point.x}</b>: {point.y:.2f}%",
            "hide_empty": True,
            "yPlotBands": [
                {"from": 20, "to": 100, "color": "#e6c3c3"},
                {"from": 5, "to": 20, "color": "#e6dcc3"},
                {"from": 0, "to": 5, "color": "#c3e6c3"},
            ],
        }

        self.add_section(
            name="Adapter Content",
            anchor="dragen_fastqc_adapter_content",
            description="""The cumulative percentage count of the proportion of your
            library which has seen each of the adapter sequences at each position.""",
            helptext="""
            Note that only samples with ≥ 0.1% adapter contamination are shown.

            There may be several lines per sample, as one is shown for each adapter
            detected in the file.

            From the [FastQC Help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/10%20Adapter%20Content.html):

            _The plot shows a cumulative percentage count of the proportion
            of your library which has seen each of the adapter sequences at each position.
            Once a sequence has been seen in a read it is counted as being present
            right through to the end of the read so the percentages you see will only
            increase as the read length goes on._
            """,
            plot=linegraph.plot(data, pconfig),
        )



""" MultiQC submodule to parse output from RSeQC inner_distance.py
http://rseqc.sourceforge.net/#inner-distance-py """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC inner_distance frequency reports and parse their data"""

    # Set up vars
    self.inner_distance = dict()
    self.inner_distance_pct = dict()

    # Go through files and parse data
    for f in self.find_log_files("rseqc/inner_distance"):
        if f["s_name"] in self.inner_distance:
            log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
        self.add_data_source(f, section="inner_distance")
        # saving to temporary variable for SE checking later
        parsed_data = dict()
        for line in f["f"].splitlines():
            s = line.split()
            try:
                avg_pos = (float(s[0]) + float(s[1])) / 2.0
                parsed_data[avg_pos] = float(s[2])
            except Exception:
                # Don't bother running through whole file if wrong
                break

        # Only add if we actually found something i,e it was PE data
        if len(parsed_data) > 0:
            self.inner_distance[f["s_name"]] = parsed_data

    # Filter to strip out ignored sample names
    self.inner_distance = self.ignore_samples(self.inner_distance)

    if len(self.inner_distance) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None, f["s_name"])

    # Write data to file
    self.write_data_file(self.inner_distance, "rseqc_inner_distance")

    # Make a normalised percentage version of the data
    for s_name in self.inner_distance:
        self.inner_distance_pct[s_name] = dict()
        total = sum(self.inner_distance[s_name].values())
        if total == 0:
            continue
        for k, v in self.inner_distance[s_name].items():
            self.inner_distance_pct[s_name][k] = (v / total) * 100

    # Add line graph to section
    pconfig = {
        "id": "rseqc_inner_distance_plot",
        "title": "RSeQC: Inner Distance",
        "ylab": "Counts",
        "xlab": "Inner Distance (bp)",
        "tt_label": "<strong>{point.x} bp</strong>: {point.y:.2f}",
        "data_labels": [{"name": "Counts", "ylab": "Counts"}, {"name": "Percentages", "ylab": "Percentage"}],
    }
    self.add_section(
        name="Inner Distance",
        anchor="rseqc-inner_distance",
        description='<a href="http://rseqc.sourceforge.net/#inner-distance-py" target="_blank">Inner Distance</a>'
        " calculates the inner distance"
        " (or insert size) between two paired RNA reads."
        " Note that this can be negative if fragments overlap.",
        plot=linegraph.plot([self.inner_distance, self.inner_distance_pct], pconfig),
    )

    # Return number of samples found
    return len(self.inner_distance)



#!/usr/bin/env python

""" MultiQC report module. Holds the output from each
module. Is available to subsequent modules. Contains
helper functions to generate markup for report. """


import fnmatch
import inspect
import io
import json
import mimetypes
import os
import re
import time
from collections import defaultdict, OrderedDict
from pathlib import Path
import rich
import rich.progress
import yaml
from yaml.representer import Representer

from multiqc.utils import lzstring

from . import config

logger = config.logger

# Treat defaultdict and OrderedDict as normal dicts for YAML output
yaml.add_representer(defaultdict, Representer.represent_dict)
yaml.add_representer(OrderedDict, Representer.represent_dict)


# Set up global variables shared across modules
# Inside a function so that the global vars are reset if MultiQC is run more than once within a single session / environment
def init():
    global general_stats_data
    general_stats_data = list()

    global general_stats_headers
    general_stats_headers = list()

    global general_stats_html
    general_stats_html = ""

    global data_sources
    data_sources = defaultdict(lambda: defaultdict(lambda: defaultdict()))

    global plot_data
    plot_data = dict()

    global html_ids
    html_ids = list()

    global lint_errors
    lint_errors = list()

    global num_hc_plots
    num_hc_plots = 0

    global num_mpl_plots
    num_mpl_plots = 0

    global saved_raw_data
    saved_raw_data = dict()

    global last_found_file
    last_found_file = None

    global runtimes
    runtimes = {
        "total": 0,
        "total_sp": 0,
        "total_mods": 0,
        "total_compression": 0,
        "sp": defaultdict(),
        "mods": defaultdict(),
    }

    global file_search_stats
    file_search_stats = {
        "skipped_symlinks": 0,
        "skipped_not_a_file": 0,
        "skipped_ignore_pattern": 0,
        "skipped_filesize_limit": 0,
        "skipped_module_specific_max_filesize": 0,
        "skipped_no_match": 0,
        "skipped_directory_fn_ignore_dirs": 0,
        "skipped_file_contents_search_errors": 0,
    }

    global searchfiles
    searchfiles = list()

    # Make a dict of discovered files for each seach key
    global files
    files = dict()

    # Map of software tools to a set of unique version strings
    global software_versions
    software_versions = defaultdict(lambda: defaultdict(list))


def is_searching_in_source_dir(path: Path) -> bool:
    """
    Checks whether MultiQC is searching for files in the source code folder
    """
    multiqc_installation_dir_files = [
        "LICENSE",
        "CHANGELOG.md",
        "Dockerfile",
        "MANIFEST.in",
        ".gitmodules",
        "README.md",
        "CSP.txt",
        "setup.py",
        ".gitignore",
    ]

    filenames = [f.name for f in path.iterdir() if f.is_file()]

    if len(filenames) > 0 and all([fn in filenames for fn in multiqc_installation_dir_files]):
        logger.error(f"Error: MultiQC is running in source code directory! {path}")
        logger.warning("Please see the docs for how to use MultiQC: https://multiqc.info/docs/#running-multiqc")
        return True
    else:
        return False


def handle_analysis_path(item: Path):
    """
    Branching logic to handle analysis paths (directories and files)
    Walks directory trees recursively calling pathlib's `path.iterdir()`.
    Guaranteed to work correctly with symlinks even on non-POSIX compliant filesystems.
    """
    if item.is_symlink() and config.ignore_symlinks:
        file_search_stats["skipped_symlinks"] += 1
        return
    elif item.is_file():
        searchfiles.append([item.name, os.fspath(item.parent)])
    elif item.is_dir():
        # Skip directory if it matches ignore patterns
        d_matches = any(d for d in config.fn_ignore_dirs if item.match(d.rstrip(os.sep)))
        p_matches = any(p for p in config.fn_ignore_paths if item.match(p.rstrip(os.sep)))
        if d_matches or p_matches:
            file_search_stats["skipped_directory_fn_ignore_dirs"] += 1
            return

        # Check not running in install directory
        if is_searching_in_source_dir(item):
            return

        for item in item.iterdir():
            handle_analysis_path(item)


def get_filelist(run_module_names):
    """
    Go through all supplied search directories and assembly a master
    list of files to search. Then fire search functions for each file.
    """
    # Prep search patterns
    spatterns = [{}, {}, {}, {}, {}, {}, {}]
    runtimes["sp"] = defaultdict()
    ignored_patterns = []
    skipped_patterns = []
    for key, sps in config.sp.items():
        mod_name = key.split("/", 1)[0]
        if mod_name.lower() not in [m.lower() for m in run_module_names]:
            ignored_patterns.append(key)
            continue
        files[key] = list()
        if not isinstance(sps, list):
            sps = [sps]

        # Warn if we have any unrecognised search pattern keys
        expected_sp_keys = [
            "fn",
            "fn_re",
            "contents",
            "contents_re",
            "num_lines",
            "shared",
            "skip",
            "max_filesize",
            "exclude_fn",
            "exclude_fn_re",
            "exclude_contents",
            "exclude_contents_re",
        ]
        unrecognised_keys = [y for x in sps for y in x.keys() if y not in expected_sp_keys]
        if len(unrecognised_keys) > 0:
            logger.warning(f"Unrecognised search pattern keys for '{key}': {', '.join(unrecognised_keys)}")

        # Check if we are skipping this search key
        if any([x.get("skip") for x in sps]):
            skipped_patterns.append(key)

        # Split search patterns according to speed of execution.
        if any([x for x in sps if "contents_re" in x]):
            if any([x for x in sps if "num_lines" in x]):
                spatterns[4][key] = sps
            elif any([x for x in sps if "max_filesize" in x]):
                spatterns[5][key] = sps
            else:
                spatterns[6][key] = sps
        elif any([x for x in sps if "contents" in x]):
            if any([x for x in sps if "num_lines" in x]):
                spatterns[1][key] = sps
            elif any([x for x in sps if "max_filesize" in x]):
                spatterns[2][key] = sps
            else:
                spatterns[3][key] = sps
        else:
            spatterns[0][key] = sps

    if len(ignored_patterns) > 0:
        logger.debug(f"Ignored {len(ignored_patterns)} search patterns as didn't match running modules.")

    if len(skipped_patterns) > 0:
        logger.info(f"Skipping {len(skipped_patterns)} file search patterns")
        logger.debug(f"Skipping search patterns: {', '.join(skipped_patterns)}")

    def add_file(fn, root):
        """
        Function applied to each file found when walking the analysis
        directories. Runs through all search patterns and returns True
        if a match is found.
        """
        f = {"fn": fn, "root": root}

        # Check that this is a file and not a pipe or anything weird
        if not os.path.isfile(os.path.join(root, fn)):
            file_search_stats["skipped_not_a_file"] += 1
            return False

        # Check that we don't want to ignore this file
        i_matches = [n for n in config.fn_ignore_files if fnmatch.fnmatch(fn, n)]
        if len(i_matches) > 0:
            file_search_stats["skipped_ignore_pattern"] += 1
            return False

        # Limit search to small files, to avoid 30GB FastQ files etc.
        try:
            f["filesize"] = os.path.getsize(os.path.join(root, fn))
        except (IOError, OSError, ValueError, UnicodeDecodeError):
            logger.debug(f"Couldn't read file when checking filesize: {fn}")
        else:
            if f["filesize"] > config.log_filesize_limit:
                file_search_stats["skipped_filesize_limit"] += 1
                return False

        # Use mimetypes to exclude binary files where possible
        if not re.match(r".+_mqc\.(png|jpg|jpeg)", f["fn"]) and config.ignore_images:
            (ftype, encoding) = mimetypes.guess_type(os.path.join(f["root"], f["fn"]))
            if encoding is not None:
                return False
            if ftype is not None and ftype.startswith("image"):
                return False

        # Test file for each search pattern
        file_matched = False
        for patterns in spatterns:
            for key, sps in patterns.items():
                start = time.time()
                for sp in sps:
                    if search_file(sp, f, key):
                        # Check that we shouldn't exclude this file
                        if not exclude_file(sp, f):
                            # Looks good! Remember this file
                            files[key].append(f)
                            file_search_stats[key] = file_search_stats.get(key, 0) + 1
                            file_matched = True
                        # Don't keep searching this file for other modules
                        if not sp.get("shared", False) and key not in config.filesearch_file_shared:
                            runtimes["sp"][key] = runtimes["sp"].get(key, 0) + (time.time() - start)
                            return True
                        # Don't look at other patterns for this module
                        break
                runtimes["sp"][key] = runtimes["sp"].get(key, 0) + (time.time() - start)

        return file_matched

    # Go through the analysis directories and get file list
    total_sp_starttime = time.time()
    for path in config.analysis_dir:
        handle_analysis_path(Path(path))

    # Search through collected files
    console = rich.console.Console(
        stderr=True,
        highlight=False,
        force_interactive=False if config.no_ansi else None,
        color_system=None if config.no_ansi else "auto",
    )
    progress_obj = rich.progress.Progress(
        "[blue]|[/]      ",
        rich.progress.SpinnerColumn(),
        "[blue]{task.description}[/] |",
        rich.progress.BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        "[green]{task.completed}/{task.total}",
        "[dim]{task.fields[s_fn]}",
        console=console,
        disable=config.no_ansi or config.quiet,
    )
    with progress_obj as progress:
        mqc_task = progress.add_task("searching", total=len(searchfiles), s_fn="")
        for sf in searchfiles:
            progress.update(mqc_task, advance=1, s_fn=os.path.join(sf[1], sf[0])[-50:])
            if not add_file(sf[0], sf[1]):
                file_search_stats["skipped_no_match"] += 1
        progress.update(mqc_task, s_fn="")

    runtimes["total_sp"] = time.time() - total_sp_starttime
    if config.profile_runtime:
        logger.info(f"Profile-runtime: Searching files took {runtimes['total_sp']:.2f}s")

    # Debug log summary about what we skipped
    summaries = []
    for key in sorted(file_search_stats, key=file_search_stats.get, reverse=True):
        if "skipped_" in key and file_search_stats[key] > 0:
            summaries.append(f"{key}: {file_search_stats[key]}")
    logger.debug(f"Summary of files that were skipped by the search: [{'] // ['.join(summaries)}]")


def search_file(pattern, f, module_key):
    """
    Function to searach a single file for a single search pattern.
    """

    global file_search_stats
    fn_matched = False
    contents_matched = False

    # Search pattern specific filesize limit
    if pattern.get("max_filesize") is not None and "filesize" in f:
        if f["filesize"] > pattern.get("max_filesize"):
            file_search_stats["skipped_module_specific_max_filesize"] += 1
            return False

    # Search by file name (glob)
    if pattern.get("fn") is not None:
        if fnmatch.fnmatch(f["fn"], pattern["fn"]):
            fn_matched = True
            if pattern.get("contents") is None and pattern.get("contents_re") is None:
                return True
        else:
            return False

    # Search by file name (regex)
    if pattern.get("fn_re") is not None:
        if re.match(pattern["fn_re"], f["fn"]):
            fn_matched = True
            if pattern.get("contents") is None and pattern.get("contents_re") is None:
                return True
        else:
            return False

    # Search by file contents
    if pattern.get("contents") is not None or pattern.get("contents_re") is not None:
        if pattern.get("contents_re") is not None:
            repattern = re.compile(pattern["contents_re"])
        if "contents_lines" not in f or ("num_lines" in pattern and len(f["contents_lines"]) < pattern["num_lines"]):
            f["contents_lines"] = []
            file_path = os.path.join(f["root"], f["fn"])

            try:
                fh = io.open(file_path, "r", encoding="utf-8")
            except Exception as e:
                if config.report_readerrors:
                    logger.debug(f"Couldn't read file when looking for output: {file_path}, {e}")
                file_search_stats["skipped_file_contents_search_errors"] += 1
                return False
            else:
                try:
                    for i, line in enumerate(fh):
                        f["contents_lines"].append(line)
                        if i >= config.filesearch_lines_limit and i >= pattern.get("num_lines", 0):
                            break
                except UnicodeDecodeError as e:
                    if config.report_readerrors:
                        logger.debug(
                            f"Couldn't read file as a utf-8 text when looking for output: {file_path}, {e}. "
                            f"Usually because it's a binary file. But sometimes there are single non-unicode "
                            f"characters, so attempting reading while skipping such characters."
                        )
                    try:
                        with io.open(file_path, "r", encoding="utf-8", errors="ignore") as fh:
                            for i, line in enumerate(fh):
                                f["contents_lines"].append(line)
                                if i >= config.filesearch_lines_limit and i >= pattern.get("num_lines", 0):
                                    break
                    except Exception as e:
                        if config.report_readerrors:
                            logger.debug(f"Still couldn't read the file, skipping: {file_path}, {e}")
                        file_search_stats["skipped_file_contents_search_errors"] += 1
                        return False
                    else:
                        if not f["contents_lines"]:
                            if config.report_readerrors:
                                logger.debug(f"No utf-8 lines were read from the file, skipping {file_path}")
                            file_search_stats["skipped_file_contents_search_errors"] += 1
                            return False
                except Exception as e:
                    if config.report_readerrors:
                        logger.debug(f"Couldn't read file when looking for output: {file_path}, {e}")
                    file_search_stats["skipped_file_contents_search_errors"] += 1
                    return False
            finally:
                fh.close()

        # Go through the parsed file contents
        for i, line in enumerate(f["contents_lines"]):
            # Break if we've searched enough lines for this pattern
            if pattern.get("num_lines") and i >= pattern.get("num_lines"):
                break

            # Search by file contents (string)
            if pattern.get("contents") is not None:
                if pattern["contents"] in line:
                    contents_matched = True
                    if pattern.get("fn") is None and pattern.get("fn_re") is None:
                        return True
                    break
            # Search by file contents (regex)
            elif pattern.get("contents_re") is not None:
                if re.search(repattern, line):
                    contents_matched = True
                    if pattern.get("fn") is None and pattern.get("fn_re") is None:
                        return True
                    break

    return fn_matched and contents_matched


def exclude_file(sp, f):
    """
    Exclude discovered files if they match the special exclude_
    search pattern keys
    """
    # Make everything a list if it isn't already
    for k in sp:
        if k in ["exclude_fn", "exclude_fn_re" "exclude_contents", "exclude_contents_re"]:
            if not isinstance(sp[k], list):
                sp[k] = [sp[k]]

    # Search by file name (glob)
    if "exclude_fn" in sp:
        for pat in sp["exclude_fn"]:
            if fnmatch.fnmatch(f["fn"], pat):
                return True

    # Search by file name (regex)
    if "exclude_fn_re" in sp:
        for pat in sp["exclude_fn_re"]:
            if re.match(pat, f["fn"]):
                return True

    # Search the contents of the file
    if "exclude_contents" in sp or "exclude_contents_re" in sp:
        # Compile regex patterns if we have any
        if "exclude_contents_re" in sp:
            sp["exclude_contents_re"] = [re.compile(pat) for pat in sp["exclude_contents_re"]]
        with io.open(os.path.join(f["root"], f["fn"]), "r", encoding="utf-8") as fh:
            for line in fh:
                if "exclude_contents" in sp:
                    for pat in sp["exclude_contents"]:
                        if pat in line:
                            return True
                if "exclude_contents_re" in sp:
                    for pat in sp["exclude_contents_re"]:
                        if re.search(pat, line):
                            return True
    return False


def data_sources_tofile():
    fn = f"multiqc_sources.{config.data_format_extensions[config.data_format]}"
    with io.open(os.path.join(config.data_dir, fn), "w", encoding="utf-8") as f:
        if config.data_format == "json":
            jsonstr = json.dumps(data_sources, indent=4, ensure_ascii=False)
            print(jsonstr.encode("utf-8", "ignore").decode("utf-8"), file=f)
        elif config.data_format == "yaml":
            yaml.dump(data_sources, f, default_flow_style=False)
        else:
            lines = [["Module", "Section", "Sample Name", "Source"]]
            for mod in data_sources:
                for sec in data_sources[mod]:
                    for s_name, source in data_sources[mod][sec].items():
                        lines.append([mod, sec, s_name, source])
            body = "\n".join(["\t".join(line) for line in lines])
            print(body.encode("utf-8", "ignore").decode("utf-8"), file=f)


def dois_tofile(modules_output):
    """Find all DOIs listed in report sections and write to a file"""
    # Collect DOIs
    dois = {"MultiQC": ["10.1093/bioinformatics/btw354"]}
    for mod in modules_output:
        if mod.doi is not None and mod.doi != "" and mod.doi != []:
            dois[mod.anchor] = mod.doi
    # Write to a file
    fn = f"multiqc_citations.{config.data_format_extensions[config.data_format]}"
    with io.open(os.path.join(config.data_dir, fn), "w", encoding="utf-8") as f:
        if config.data_format == "json":
            jsonstr = json.dumps(dois, indent=4, ensure_ascii=False)
            print(jsonstr.encode("utf-8", "ignore").decode("utf-8"), file=f)
        elif config.data_format == "yaml":
            yaml.dump(dois, f, default_flow_style=False)
        else:
            body = ""
            for mod, dois in dois.items():
                for doi in dois:
                    body += f"{doi}{' ' * (50 - len(doi))} # {mod}\n"
            print(body.encode("utf-8", "ignore").decode("utf-8"), file=f)


def save_htmlid(html_id, skiplint=False):
    """Take a HTML ID, sanitise for HTML, check for duplicates and save.
    Returns sanitised, unique ID"""
    global html_ids
    global lint_errors

    # Trailing whitespace
    html_id_clean = html_id.strip()

    # Trailing underscores
    html_id_clean = html_id_clean.strip("_")

    # Must begin with a letter
    if re.match(r"^[a-zA-Z]", html_id_clean) is None:
        html_id_clean = f"mqc_{html_id_clean}"

    # Replace illegal characters
    html_id_clean = re.sub("[^a-zA-Z0-9_-]+", "_", html_id_clean)

    # Validate if linting
    if config.strict and not skiplint:
        modname = ""
        codeline = ""
        callstack = inspect.stack()
        for n in callstack:
            if "multiqc/modules/" in n[1] and "base_module.py" not in n[1]:
                callpath = n[1].split("multiqc/modules/", 1)[-1]
                modname = f">{callpath}< "
                codeline = n[4][0].strip()
                break
    if config.strict and not skiplint and html_id != html_id_clean:
        errmsg = f"LINT: {modname}HTML ID was not clean ('{html_id}' -> '{html_id_clean}') ## {codeline}"
        logger.error(errmsg)
        lint_errors.append(errmsg)

    # Check for duplicates
    i = 1
    html_id_base = html_id_clean
    while html_id_clean in html_ids:
        html_id_clean = f"{html_id_base}-{i}"
        i += 1
        if config.strict and not skiplint:
            errmsg = f"LINT: {modname}HTML ID was a duplicate ({html_id_clean}) ## {codeline}"
            logger.error(errmsg)
            lint_errors.append(errmsg)

    # Remember and return
    html_ids.append(html_id_clean)
    return html_id_clean


def compress_json(data):
    """Take a Python data object. Convert to JSON and compress using lzstring"""
    json_string = json.dumps(data).encode("utf-8", "ignore").decode("utf-8")
    json_string = sanitise_json(json_string)
    x = lzstring.LZString()
    return x.compressToBase64(json_string)


def sanitise_json(json_string):
    """
    The Python json module uses a bunch of values which are valid JavaScript
    but invalid JSON. These crash the browser when parsing the JSON.
    Nothing in the MultiQC front-end uses these values, so instead we just
    do a find-and-replace for them and switch them with `null`, which works fine.

    Side effect: Any string values that include the word "Infinity"
    (case-sensitive) will have it switched for "null". Hopefully that doesn't happen
    a lot, otherwise we'll have to do this in a more complicated manner.
    """
    json_string = re.sub(r"\bNaN\b", "null", json_string)
    json_string = re.sub(r"\b-?Infinity\b", "null", json_string)
    return json_string



""" MultiQC functions to plot a scatter plot """

import logging

from multiqc.utils import config
from multiqc.plots.plotly import scatter

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, pconfig=None):
    """Plot a scatter plot with X,Y data.
    :param data: 2D dict, first keys as sample names, then x:y data pairs
    :param pconfig: optional dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML and JS, ready to be inserted into the page
    """
    if pconfig is None:
        pconfig = {}

    # Allow user to overwrite any given config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for point, v in config.custom_plot_config[pconfig["id"]].items():
            pconfig[point] = v

    # Given one dataset - turn it into a list
    if not isinstance(data, list):
        data = [data]

    # Generate the data dict structure expected by HighCharts series
    plotdata = list()
    for data_index, ds in enumerate(data):
        d = list()
        for s_name in ds:
            # Ensure any overwriting conditionals from data_labels (e.g. ymax) are taken in consideration
            series_config = pconfig.copy()
            if "data_labels" in pconfig and isinstance(
                pconfig["data_labels"][data_index], dict
            ):  # if not a dict: only dataset name is provided
                series_config.update(pconfig["data_labels"][data_index])

            if not isinstance(ds[s_name], list):
                ds[s_name] = [ds[s_name]]
            for point in ds[s_name]:
                if point["x"] is not None:
                    if "xmax" in series_config and float(point["x"]) > float(series_config["xmax"]):
                        continue
                    if "xmin" in series_config and float(point["x"]) < float(series_config["xmin"]):
                        continue
                if point["y"] is not None:
                    if "ymax" in series_config and float(point["y"]) > float(series_config["ymax"]):
                        continue
                    if "ymin" in series_config and float(point["y"]) < float(series_config["ymin"]):
                        continue
                if "name" in point:
                    point["name"] = f'{s_name}: {point["name"]}'
                else:
                    point["name"] = s_name

                for k in ["color", "opacity", "marker_size", "marker_line_width"]:
                    if k not in point and k in series_config:
                        v = series_config[k]
                        if isinstance(v, dict) and s_name in v:
                            point[k] = v[s_name]
                        else:
                            point[k] = v
                d.append(point)
        plotdata.append(d)

    if pconfig.get("square"):
        # Making sure HighCharts doesn't get creative in adding different paddings
        pconfig["endOnTick"] = False
        if "ymax" not in pconfig and "xmax" not in pconfig:
            # Find the max value
            max_val = 0
            for d in plotdata:
                for s in d:
                    max_val = max(max_val, s["x"], s["y"])
            max_val = 1.02 * max_val  # add 2% padding
            pconfig["xmax"] = pconfig.get("xmax", max_val)
            pconfig["ymax"] = pconfig.get("ymax", max_val)

    # Add on annotation data series
    try:
        if pconfig.get("extra_series"):
            extra_series = pconfig["extra_series"]
            if isinstance(pconfig["extra_series"], dict):
                extra_series = [[pconfig["extra_series"]]]
            elif isinstance(pconfig["extra_series"], list) and isinstance(pconfig["extra_series"][0], dict):
                extra_series = [pconfig["extra_series"]]
            for i, es in enumerate(extra_series):
                for s in es:
                    plotdata[i].append(s)
    except Exception:
        pass

    # Make a plot
    mod = get_template_mod()
    if "scatter" in mod.__dict__ and callable(mod.scatter):
        try:
            return mod.scatter(plotdata, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return scatter.plot(plotdata, pconfig)



""" MultiQC functions to plot a linegraph """

import inspect
import logging
import re
from typing import List, Dict

from multiqc.utils import config, mqc_colour, report
from multiqc.plots.plotly import line

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, pconfig=None):
    """Plot a line graph with X,Y data.
    :param data: 2D dict, first keys as sample names, then x:y data pairs
    :param pconfig: optional dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML and JS, ready to be inserted into the page
    """
    # Don't just use {} as the default argument as it's mutable. See:
    # http://python-guide-pt-br.readthedocs.io/en/latest/writing/gotchas/
    if pconfig is None:
        pconfig = {}

    # Allow user to overwrite any given config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for k, v in config.custom_plot_config[pconfig["id"]].items():
            pconfig[k] = v

    # Given one dataset - turn it into a list
    if not isinstance(data, list):
        data = [data]

    # Validate config if linting
    if config.strict:
        # Get module name
        modname = ""
        callstack = inspect.stack()
        for n in callstack:
            if "multiqc/modules/" in n[1] and "base_module.py" not in n[1]:
                callpath = n[1].split("multiqc/modules/", 1)[-1]
                modname = f">{callpath}< "
                break
        # Look for essential missing pconfig keys
        for k in ["id", "title", "ylab"]:
            if k not in pconfig:
                errmsg = f"LINT: {modname}Linegraph pconfig was missing key '{k}'"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)
        # Check plot title format
        if not re.match(r"^[^:]*\S: \S[^:]*$", pconfig.get("title", "")):
            errmsg = "LINT: {} Linegraph title did not match format 'Module: Plot Name' (found '{}')".format(
                modname, pconfig.get("title", "")
            )
            logger.error(errmsg)
            report.lint_errors.append(errmsg)

    # Smooth dataset if requested in config
    if pconfig.get("smooth_points", None) is not None:
        for i, d in enumerate(data):
            data[i] = smooth_line_data(d, pconfig["smooth_points"])

    # Add sane plotting config defaults
    for idx, yp in enumerate(pconfig.get("yPlotLines", [])):
        pconfig["yPlotLines"][idx]["width"] = pconfig["yPlotLines"][idx].get("width", 2)

    # Add initial axis labels if defined in `data_labels` but not main config
    if pconfig.get("ylab") is None:
        try:
            pconfig["ylab"] = pconfig["data_labels"][0]["ylab"]
        except Exception:
            pass
    if pconfig.get("xlab") is None:
        try:
            pconfig["xlab"] = pconfig["data_labels"][0]["xlab"]
        except Exception:
            pass

    # Generate the data dict structure expected by HighCharts series
    plotdata: List[List[Dict]] = []
    for data_index, d in enumerate(data):
        thisplotdata: List[Dict] = []

        # Ensure any overwritten conditionals from data_labels (e.g. ymax or categories) are taken in consideration
        dataset_config = pconfig.copy()
        if "data_labels" in pconfig and isinstance(
            pconfig["data_labels"][data_index], dict
        ):  # if not a dict: only dataset name is provided
            dataset_config.update(pconfig["data_labels"][data_index])

        if "categories" in dataset_config:
            if not isinstance(pconfig["categories"], list):
                dataset_config["categories"] = list()

            # Add any new categories
            for s in sorted(d.keys()):
                for k in d[s].keys():
                    if k not in dataset_config["categories"]:
                        dataset_config["categories"].append(k)

            # Save adjusted categories per-dataset
            del pconfig["categories"]
            if "data_labels" not in pconfig:
                pconfig["data_labels"] = [{}] * len(data)
            pconfig["data_labels"] = [({"name": dl} if isinstance(dl, str) else dl) for dl in pconfig["data_labels"]]
            pconfig["data_labels"][data_index]["categories"] = dataset_config["categories"]

        for s in sorted(d.keys()):
            # Ensure any overwritten conditionals from data_labels (e.g. ymax) are taken in consideration
            series_config = dataset_config.copy()
            pairs = []
            maxval = 0
            if "categories" in series_config:
                # Go through categories and add either data or a blank
                for k in series_config["categories"]:
                    try:
                        pairs.append(d[s][k])
                        maxval = max(maxval, d[s][k])
                    except KeyError:
                        pairs.append(None)
            else:
                # Discard > ymax or just hide?
                # If it never comes back into the plot, discard. If it goes above then comes back, just hide.
                discard_ymax = None
                discard_ymin = None
                for k in sorted(d[s].keys()):
                    if "xmax" in series_config and float(k) > float(series_config["xmax"]):
                        continue
                    if "xmin" in series_config and float(k) < float(series_config["xmin"]):
                        continue
                    if d[s][k] is not None and "ymax" in series_config:
                        if float(d[s][k]) > float(series_config["ymax"]):
                            discard_ymax = True
                        elif discard_ymax is True:
                            discard_ymax = False
                    if d[s][k] is not None and "ymin" in series_config:
                        if float(d[s][k]) > float(series_config["ymin"]):
                            discard_ymin = True
                        elif discard_ymin is True:
                            discard_ymin = False

                # Build the plot data structure
                for k in sorted(d[s].keys()):
                    if k is not None:
                        if "xmax" in series_config and float(k) > float(series_config["xmax"]):
                            continue
                        if "xmin" in series_config and float(k) < float(series_config["xmin"]):
                            continue
                    if d[s][k] is not None:
                        if (
                            "ymax" in series_config
                            and float(d[s][k]) > float(series_config["ymax"])
                            and discard_ymax is not False
                        ):
                            continue
                        if (
                            "ymin" in series_config
                            and float(d[s][k]) < float(series_config["ymin"])
                            and discard_ymin is not False
                        ):
                            continue
                    pairs.append([k, d[s][k]])
                    try:
                        maxval = max(maxval, d[s][k])
                    except TypeError:
                        pass
            if maxval > 0 or series_config.get("hide_empty") is not True:
                this_series: Dict = {"name": s, "data": pairs}
                try:
                    this_series["color"] = series_config["colors"][s]
                except Exception:
                    pass
                thisplotdata.append(this_series)
        plotdata.append(thisplotdata)

    # Add on annotation data series
    try:
        if pconfig.get("extra_series"):
            extra_series = pconfig["extra_series"]
            if isinstance(pconfig["extra_series"], dict):
                extra_series = [[pconfig["extra_series"]]]
            elif isinstance(pconfig["extra_series"], list) and isinstance(pconfig["extra_series"][0], dict):
                extra_series = [pconfig["extra_series"]]
            for i, es in enumerate(extra_series):
                for s in es:
                    plotdata[i].append(s)
    except Exception:
        pass

    # Add colors to the categories if not set. Since the "plot_defaults" scale is
    # identical to default scale of the Highcharts JS library, this is not strictly
    # needed. But it future proofs when we replace Highcharts with something else.
    scale = mqc_colour.mqc_colour_scale("plot_defaults")
    for si, sd in enumerate(plotdata):
        for di, d in enumerate(sd):
            d.setdefault("color", scale.get_colour(di, lighten=1))

    # Make a plot - template custom, or interactive or flat
    mod = get_template_mod()
    if "linegraph" in mod.__dict__ and callable(mod.linegraph):
        try:
            return mod.linegraph(plotdata, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return line.plot(plotdata, pconfig)


def smooth_line_data(data: Dict[str, Dict], numpoints: int) -> Dict[str, Dict[int, int]]:
    """
    Function to take an x-y dataset and use binning to smooth to a maximum number of datapoints.
    Each datapoint in a smoothed dataset corresponds to the first point in a bin.

    Examples to show the idea:

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=6
    we want to keep the first and the last element, thus excluding the last element from the binning:
    binsize = len([0 1 2 3 4 5 6 7 8]))/(numpoints-1) = 9/5 = 1.8
    taking points in indices rounded from multiples of 1.8: [0, 1.8, 3.6, 5.4, 7.2, 9],
    ...which evaluates to first_element_in_bin_indices=[0, 2, 4, 5, 7, 9]
    picking up the elements: [0 _ 2 _ 4 5 _ 7 _ 9]

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=9
    binsize = 9/8 = 1.125
    indices: [0.0, 1.125, 2.25, 3.375, 4.5, 5.625, 6.75, 7.875, 9] -> [0, 1, 2, 3, 5, 6, 7, 8, 9]
    picking up the elements: [0 1 2 3 _ 5 6 7 8 9]

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=3
    binsize = len(d)/numpoints = 9/2 = 4.5
    indices: [0.0, 4.5, 9] -> [0, 5, 9]
    picking up the elements: [0 _ _ _ _ 5 _ _ _ 9]
    """
    smoothed_data = dict()
    for s_name, d in data.items():
        # Check that we need to smooth this data
        if len(d) <= numpoints or len(d) == 0:
            smoothed_data[s_name] = d
            continue

        binsize = (len(d) - 1) / (numpoints - 1)
        first_element_indices = [round(binsize * i) for i in range(numpoints)]
        smoothed_d = {x: y for i, (x, y) in enumerate(d.items()) if i in first_element_indices}
        smoothed_data[s_name] = smoothed_d

    return smoothed_data
